"""

Simulate localization data.

This module provides functions to simulate localization data and return
LocData objects.
Localizations are often distributed either by a spatial process of
complete-spatial randomness or following a Neyman-Scott process [1]_.
For a Neyman-Scott process parent events (representing single emitters) yield
a random number of cluster_mu events (representing localizations due to
repeated blinking). Related spatial point processes include Matérn and Thomas
processes.

Functions that are named as make_* provide point data arrays.
Functions that are named as simulate_* provide locdata.


Parts of this code is adapted from
scikit-learn/sklearn/datasets/_samples_generator.py .
(BSD 3-Clause License, Copyright (c) 2007-2020 The scikit-learn developers.)

References
----------
.. [1] Neyman, J. & Scott, E. L.,
   A Theory of the Spatial Distribution of Galaxies.
   Astrophysical Journal 1952, vol. 116, p.144.

"""

from __future__ import annotations

import logging
import sys
from collections.abc import Callable, Sequence
from itertools import chain
from typing import Any, Literal

import numpy as np
import numpy.typing as npt
import pandas as pd

from locan.data import metadata_pb2
from locan.data.locdata import LocData
from locan.data.locdata_utils import (
    _bump_property_key,
    _get_loc_property_key_per_dimension,
)
from locan.data.metadata_utils import _modify_meta
from locan.data.regions.region import (
    AxisOrientedCuboid,
    AxisOrientedHypercuboid,
    Ellipse,
    EmptyRegion,
    Interval,
    Rectangle,
    Region,
)
from locan.data.regions.region_utils import expand_region
from locan.locan_types import RandomGeneratorSeed

__all__: list[str] = [
    "make_uniform",
    "make_Poisson",
    "make_cluster",
    "make_NeymanScott",
    "make_Matern",
    "make_Thomas",
    "make_dstorm",
    "simulate_uniform",
    "simulate_Poisson",
    "simulate_cluster",
    "simulate_NeymanScott",
    "simulate_Matern",
    "simulate_Thomas",
    "simulate_dstorm",
    "simulate_tracks",
    "resample",
    "simulate_frame_numbers",
    "randomize",
]

logger = logging.getLogger(__name__)


def make_uniform(
    n_samples: int,
    region: Region | npt.ArrayLike = (0, 1),
    seed: RandomGeneratorSeed = None,
) -> npt.NDArray[np.float64]:
    """
    Provide points that are distributed by a uniform
    (complete spatial randomness) point process
    within the boundaries given by `region`.

    Parameters
    ----------
    n_samples
        The total number of localizations of the point process
    region
        The region (or support) for all features.
        If array-like it must provide upper and lower bounds for each feature.
    seed
        Random number generation seed

    Returns
    -------
    npt.NDArray[np.float64]
        The generated samples of shape (n_samples, n_features).
    """
    rng = np.random.default_rng(seed)

    if not isinstance(region, Region):
        region = Region.from_intervals(region)

    if isinstance(region, EmptyRegion):
        samples = np.array([])
    elif isinstance(
        region, (Interval, AxisOrientedCuboid, AxisOrientedHypercuboid)
    ) or (isinstance(region, Rectangle) and region.angle == 0):
        samples = rng.uniform(
            region.bounds[: region.dimension],
            region.bounds[region.dimension :],
            size=(n_samples, region.dimension),
        )
    elif isinstance(region, Ellipse) and region.width == region.height:
        radius = region.width / 2
        # angular and radial coordinates of Poisson points
        theta = rng.random(n_samples) * 2 * np.pi
        rho = radius * np.sqrt(rng.random(n_samples))
        # Convert from polar to Cartesian coordinates
        xx = rho * np.cos(theta)
        yy = rho * np.sin(theta)
        samples = np.array((xx, yy)).T
    else:
        sampling_ratio = region.region_measure / region.bounding_box.region_measure
        n_samples_updated = int(n_samples / sampling_ratio * 2)

        samples_ = []
        n_remaining = n_samples
        while n_remaining > 0:
            new_samples = rng.random(size=(n_samples_updated, region.dimension))  # type: ignore
            new_samples = region.extent * new_samples + region.bounding_box.corner  # type: ignore
            new_samples = new_samples[region.contains(new_samples)]
            samples_.append(new_samples)
            n_remaining = n_remaining - len(new_samples)

        samples = np.concatenate(samples_)
        samples = samples[0:n_samples]

    return samples


def simulate_uniform(
    n_samples: int,
    region: Region | npt.ArrayLike = (0, 1),
    seed: RandomGeneratorSeed = None,
) -> LocData:
    """
    Provide points that are distributed by a uniform Poisson point process
    within the boundaries given by `region`.

    Parameters
    ----------
    n_samples
        The total number of localizations of the point process
    region
        The region (or support) for each feature.
        If array-like it must provide upper and lower bounds for each feature.
    seed
        random number generation seed

    Returns
    -------
    LocData
        The generated samples.
    """
    parameter = locals()
    samples = make_uniform(n_samples=n_samples, region=region, seed=seed)
    region_ = region if isinstance(region, Region) else Region.from_intervals(region)
    assert region_.dimension is not None  # type narrowing # noqa: S101
    locdata = LocData.from_coordinates(coordinates=samples)
    locdata.dimension = region_.dimension
    locdata.region = region_

    # metadata
    locdata.meta.source = metadata_pb2.SIMULATION
    del locdata.meta.history[:]
    locdata.meta.history.add(name=make_uniform.__name__, parameter=str(parameter))

    return locdata


def make_Poisson(
    intensity: int | float,
    region: Region | npt.ArrayLike = (0, 1),
    seed: RandomGeneratorSeed = None,
) -> npt.NDArray[np.float64]:
    """
    Provide points that are distributed by a uniform Poisson point process
    within the boundaries given by `region`.

    Parameters
    ----------
    intensity
        The intensity (points per unit region measure) of the point process
    region
        The region (or support) for all features.
        If array-like it must provide upper and lower bounds for each feature.
    seed
        random number generation seed

    Returns
    -------
    npt.NDArray[np.float64]
        The generated samples of shape (n_samples, n_features).
    """
    rng = np.random.default_rng(seed)

    if not isinstance(region, Region):
        region = Region.from_intervals(region)

    n_samples = rng.poisson(lam=intensity * region.region_measure)

    if isinstance(region, EmptyRegion):
        samples = np.array([])
    elif n_samples == 0:
        samples = np.array([])
        assert region.dimension is not None  # type narrowing # noqa: S101
        for _i in range(region.dimension):
            samples = samples[..., np.newaxis]
    elif isinstance(
        region, (Interval, Rectangle, AxisOrientedCuboid, AxisOrientedHypercuboid)
    ):
        samples = rng.uniform(
            region.bounds[: region.dimension],
            region.bounds[region.dimension :],
            size=(n_samples, region.dimension),
        )
    elif isinstance(region, Ellipse) and region.width == region.height:
        radius = region.width / 2
        # angular and radial coordinates of Poisson points
        theta = rng.random(n_samples) * 2 * np.pi
        rho = radius * np.sqrt(rng.random(n_samples))
        # Convert from polar to Cartesian coordinates
        xx = rho * np.cos(theta)
        yy = rho * np.sin(theta)
        samples = np.array((xx, yy)).T
    else:
        sampling_ratio = region.region_measure / region.bounding_box.region_measure
        n_samples_updated = int(n_samples / sampling_ratio * 2)

        samples_ = []
        n_remaining = n_samples
        while n_remaining > 0:
            assert region.dimension is not None  # type narrowing # noqa: S101
            new_samples = rng.random(size=(n_samples_updated, region.dimension))
            new_samples = region.extent * new_samples + region.bounding_box.corner  # type: ignore
            new_samples = new_samples[region.contains(new_samples)]
            samples_.append(new_samples)
            n_remaining = n_remaining - len(new_samples)

        samples = np.concatenate(samples_)
        samples = samples[0:n_samples]

    return samples


def simulate_Poisson(
    intensity: int | float,
    region: Region | npt.ArrayLike = (0, 1),
    seed: RandomGeneratorSeed = None,
) -> LocData:
    """
    Provide points that are distributed by a uniform Poisson point process
    within the boundaries given by `region`.

    Parameters
    ----------
    intensity
        The intensity (points per unit region measure) of the point process
    region
        The region (or support) for each feature.
        If array-like it must provide upper and lower bounds for each feature.
    seed
        random number generation seed

    Returns
    -------
    LocData
        The generated samples.
    """
    parameter = locals()
    samples = make_Poisson(intensity=intensity, region=region, seed=seed)
    region_ = region if isinstance(region, Region) else Region.from_intervals(region)
    assert region_.dimension is not None  # type narrowing # noqa: S101
    locdata = LocData.from_coordinates(coordinates=samples)
    locdata.dimension = region_.dimension
    locdata.region = region_

    # metadata
    locdata.meta.source = metadata_pb2.SIMULATION
    del locdata.meta.history[:]
    locdata.meta.history.add(name=make_Poisson.__name__, parameter=str(parameter))

    return locdata


def make_cluster(
    centers: int | npt.ArrayLike = 3,
    region: Region | npt.ArrayLike = (0, 1.0),
    expansion_distance: float = 0,
    offspring: npt.ArrayLike | Callable[..., Any] | None = None,
    clip: bool = True,
    shuffle: bool = True,
    seed: RandomGeneratorSeed = None,
) -> tuple[
    npt.NDArray[np.float64], npt.NDArray[np.int64], npt.NDArray[np.float64], Region
]:
    """
    Parent positions are taken from `centers`
    or are distributed according to a homogeneous Poisson process with
    exactly `centers`
    within the boundaries given by `region` expanded by the expansion_distance.
    Each parent position is then replaced by cluster_mu offspring points as
    passed or generated by a given function.
    Offspring from parent events that are located outside the region are
    included.

    Parameters
    ----------
    centers
        The number of parents or coordinates for parent events,
        where each parent represents a cluster center.
    region
        The region (or support) for all features.
        If array-like it must provide upper and lower bounds for each feature.
    expansion_distance
        The distance by which region is expanded on all boundaries.
    offspring
        Points or function for point process to provide cluster.
        Callable must take single parent point as parameter and return an iterable.
        If array-like it must have the same length as parent events.
    clip
        If True the result will be clipped to 'region'.
        If False the extended region will be kept.
    shuffle
        If True shuffle the samples.
    seed
        random number generation seed

    Returns
    -------
    tuple[npt.NDArray[np.float64], npt.NDArray[np.int64], npt.NDArray[np.float64], Region]
       The generated samples, labels, parent_samples
       of shape (n_samples, n_features) and region
    """
    rng = np.random.default_rng(seed)

    if isinstance(region, EmptyRegion):
        samples, labels, parent_samples, region = (
            np.array([]),
            np.array([]),
            np.array([]),
            EmptyRegion(),
        )
        return samples, labels, parent_samples, region
    elif not isinstance(region, Region):
        region = Region.from_intervals(region)

    expanded_region = expand_region(region, expansion_distance)

    if isinstance(centers, (int, np.integer)):
        n_centers = int(centers)
        parent_samples = make_uniform(
            n_samples=n_centers, region=expanded_region, seed=rng
        )
    else:  # if centers is array
        parent_samples = np.array(centers)
        centers_shape = np.shape(parent_samples)
        centers_dimension = 1 if len(centers_shape) == 1 else centers_shape[1]
        if region.dimension != centers_dimension:
            raise ValueError(
                f"Region dimensions must be the same as the dimensions for each center. "
                f"Got region dimension: {region.dimension} and "
                f"center dimensions: {centers_dimension} instead."
            )
        n_centers = centers_shape[0]

    # replace parents by offspring samples
    if offspring is None:
        samples = parent_samples
        labels = np.arange(0, len(parent_samples))

    elif callable(offspring):
        try:
            offspring_samples = offspring(parent_samples)
            labels_ = [[i] * len(os) for i, os in enumerate(offspring_samples)]
        except TypeError:
            offspring_samples = []
            labels_ = []
            for i, parent in enumerate(parent_samples):
                offspring_samples_ = offspring(parent)
                offspring_samples.append(offspring_samples_)
                labels_.append([i] * len(offspring_samples_))
        samples = np.array(list(chain(*offspring_samples)))
        labels = np.array(list(chain(*labels_)))

    elif len(offspring) >= len(parent_samples):  # type: ignore
        offspring_samples = []
        labels_ = []
        for i, (os, parent) in enumerate(zip(offspring[:n_centers], parent_samples)):  # type: ignore
            if len(os) > 0:
                offspring_samples_ = np.asarray(os) + parent
                offspring_samples.append(offspring_samples_)
                labels_.append([i] * len(offspring_samples_))
        samples = np.array(list(chain(*offspring_samples)))
        labels = np.array(list(chain(*labels_)))

    else:
        raise TypeError(
            f"offspring must be callable or array-like with length >= than n_centers {n_centers}."
        )

    if (
        samples.ndim == 1
    ):  # this is to convert 1-dimensional arrays into arrays with shape (n_samples, 1).
        samples.shape = (len(samples), 1)

    if clip is True:
        if len(samples) != 0:
            inside_indices = region.contains(samples)
            samples = samples[inside_indices]
            labels = labels[inside_indices]
        region_ = region
    else:
        region_ = expanded_region

    if shuffle:
        shuffled_indices = rng.permutation(len(samples))
        samples = samples[shuffled_indices]
        labels = labels[shuffled_indices]

    if (
        len(samples) == 0
    ):  # this is to convert empty arrays into arrays with shape (n_samples, n_features).
        samples = np.array([])
        samples = samples[:, np.newaxis]

    return samples, labels, parent_samples, region_


def simulate_cluster(
    centers: int | npt.ArrayLike = 3,
    region: Region | npt.ArrayLike = (0, 1.0),
    expansion_distance: float = 0,
    offspring: npt.ArrayLike | Callable[..., Any] | None = None,
    clip: bool = True,
    shuffle: bool = True,
    seed: RandomGeneratorSeed = None,
) -> LocData:
    """
    Generate clustered point data.
    Parent positions are taken from `centers`
    or are distributed according to a homogeneous Poisson process with exactly `centers`
    within the boundaries given by `region` expanded by the expansion_distance.
    Each parent position is then replaced by offspring points as passed or generated by a given function.
    Offspring from parent events that are located outside the region are included.

    Parameters
    ----------
    centers
        The number of parents or coordinates for parent events,
        where each parent represents a cluster center.
    region
        The region (or support) for each feature.
        If array-like it must provide upper and lower bounds for each feature.
    expansion_distance
        The distance by which region is expanded on all boundaries.
    offspring
        Points or function for point process to provide cluster.
        Callable must take single parent point as parameter and return an iterable.
        If array-like it must have the same length as parent events.
    clip
        If True the result will be clipped to 'region'.
        If False the extended region will be kept.
    shuffle
        If True shuffle the samples.
    seed
        random number generation seed

    Returns
    -------
    LocData
        The generated samples.
    """
    parameter = locals()
    samples, labels, _, region = make_cluster(
        centers, region, expansion_distance, offspring, clip, shuffle, seed
    )
    region_ = region if isinstance(region, Region) else Region.from_intervals(region)  # type: ignore
    assert region_.dimension is not None  # type narrowing # noqa: S101
    locdata = LocData.from_coordinates(coordinates=samples)
    locdata.dimension = region_.dimension
    locdata.region = region_
    locdata.dataframe = locdata.dataframe.assign(cluster_label=labels)

    # metadata
    locdata.meta.source = metadata_pb2.SIMULATION
    del locdata.meta.history[:]
    locdata.meta.history.add(name=make_cluster.__name__, parameter=str(parameter))

    return locdata


def make_NeymanScott(
    parent_intensity: int | float = 100,
    region: Region | npt.ArrayLike = (0, 1.0),
    expansion_distance: float = 0,
    offspring: npt.ArrayLike | Callable[..., Any] | None = None,
    clip: bool = True,
    shuffle: bool = True,
    seed: RandomGeneratorSeed = None,
) -> tuple[
    npt.NDArray[np.float64], npt.NDArray[np.int64], npt.NDArray[np.float64], Region
]:
    """
    Generate clustered point data following a Neyman-Scott random point
    process.
    Parent positions are distributed according to a homogeneous Poisson
    process with `parent_intensity`
    within the boundaries given by `region` expanded by the expansion_distance.
    Each parent position is then replaced by offspring points as passed or
    generated by a given function.
    Offspring from parent events that are located outside the region are
    included.

    Parameters
    ----------
    parent_intensity
        The intensity (points per unit region measure) of the Poisson point
        process for parent events.
    region
        The region (or support) for all features.
        If array-like it must provide upper and lower bounds for each feature.
    expansion_distance
        The distance by which region is expanded on all boundaries.
    offspring
        Points or function for point process to provide offspring points.
        Callable must take single parent point as parameter.
        If array-like it must have enough elements to fit the randomly
        generated number of parent events.
    clip
        If True the result will be clipped to 'region'.
        If False the extended region will be kept.
    shuffle
        If True shuffle the samples.
    seed
        random number generation seed

    Returns
    -------
    tuple[npt.NDArray[np.float64], npt.NDArray[np.int64], npt.NDArray[np.float64], Region]
       The generated samples, labels, parent_samples of shape
       (n_samples, n_features) and region
    """
    rng = np.random.default_rng(seed)

    if isinstance(region, EmptyRegion):
        samples, labels, parent_samples, region = (
            np.array([]),
            np.array([]),
            np.array([]),
            EmptyRegion(),
        )
        return samples, labels, parent_samples, region
    elif not isinstance(region, Region):
        region = Region.from_intervals(region)

    expanded_region = expand_region(region, expansion_distance)

    parent_samples = make_Poisson(
        intensity=parent_intensity, region=expanded_region, seed=rng
    )

    # replace parents by offspring samples
    if offspring is None:
        samples = parent_samples
        labels = np.arange(0, len(parent_samples))

    elif callable(offspring):
        try:
            offspring_samples = offspring(parent_samples)
            labels_ = [[i] * len(os) for i, os in enumerate(offspring_samples)]
        except TypeError:
            offspring_samples = []
            labels_ = []
            for i, parent in enumerate(parent_samples):
                offspring_samples_ = offspring(parent)
                offspring_samples.append(offspring_samples_)
                labels_.append([i] * len(offspring_samples_))
        samples = np.array(list(chain(*offspring_samples)))
        labels = np.array(list(chain(*labels_)))

    elif len(offspring) >= len(parent_samples):  # type: ignore
        offspring_samples = []
        labels_ = []
        if isinstance(offspring, np.ndarray):
            offspring_samples = (
                np.asarray(offspring[: len(parent_samples)]) + parent_samples
            )
            labels_ = [[i] * len(os) for i, os in enumerate(offspring_samples)]
        else:
            for i, (os, parent) in enumerate(
                zip(offspring[: len(parent_samples)], parent_samples)  # type: ignore
            ):
                if len(os) > 0:
                    offspring_samples_ = np.asarray(os) + parent
                    offspring_samples.append(offspring_samples_)
                    labels_.append([i] * len(offspring_samples_))
        samples = np.array(list(chain(*offspring_samples)))
        labels = np.array(list(chain(*labels_)))

    else:
        raise TypeError(
            f"offspring must be callable or array-like with "
            f"length >= n_centers {len(parent_samples)}."
        )

    if (
        samples.ndim == 1
    ):  # this is to convert 1-dimensional arrays into arrays with shape (n_samples, 1).
        samples.shape = (len(samples), 1)

    if clip is True:
        if len(samples) != 0:
            inside_indices = region.contains(samples)
            samples = samples[inside_indices]
            labels = labels[inside_indices]
        region_ = region
    else:
        region_ = expanded_region

    if shuffle:
        shuffled_indices = rng.permutation(len(samples))
        samples = samples[shuffled_indices]
        labels = labels[shuffled_indices]

    if (
        len(samples) == 0
    ):  # this is to convert empty arrays into arrays with shape (n_samples, n_features).
        samples = np.array([])
        samples = samples[:, np.newaxis]

    return samples, labels, parent_samples, region_


def simulate_NeymanScott(
    parent_intensity: int | float = 100,
    region: Region | npt.ArrayLike = (0, 1.0),
    expansion_distance: float = 0,
    offspring: npt.ArrayLike | Callable[..., Any] | None = None,
    clip: bool = True,
    shuffle: bool = True,
    seed: RandomGeneratorSeed = None,
) -> LocData:
    """
    Generate clustered point data following a Neyman-Scott random point
    process.
    Parent positions are distributed according to a homogeneous Poisson
    process with `parent_intensity`
    within the boundaries given by `region` expanded by the expansion_distance.
    Each parent position is then replaced by offspring points as passed or
    generated by a given function.
    Offspring from parent events that are located outside the region are
    included.

    Parameters
    ----------
    parent_intensity
        The intensity (points per unit region measure) of the Poisson point
        process for parent events.
    region
        The region (or support) for each feature.
        If array-like it must provide upper and lower bounds for each feature.
    expansion_distance
        The distance by which region is expanded on all boundaries.
    offspring
        Points or function for point process to provide offspring points.
        Callable must take single parent point as parameter.
        If array-like it must have enough elements to fit the randomly
        generated number of parent events.
    clip
        If True the result will be clipped to 'region'.
        If False the extended region will be kept.
    shuffle
        If True shuffle the samples.
    seed
        random number generation seed

    Returns
    -------
    LocData
        The generated samples.
    """
    parameter = locals()
    samples, labels, _, region = make_NeymanScott(
        parent_intensity, region, expansion_distance, offspring, clip, shuffle, seed
    )
    region_ = region if isinstance(region, Region) else Region.from_intervals(region)  # type: ignore
    assert region_.dimension is not None  # type narrowing # noqa: S101
    locdata = LocData.from_coordinates(coordinates=samples)
    locdata.dimension = region_.dimension
    locdata.region = region_
    locdata.dataframe = locdata.dataframe.assign(cluster_label=labels)

    # metadata
    locdata.meta.source = metadata_pb2.SIMULATION
    del locdata.meta.history[:]
    locdata.meta.history.add(name=make_NeymanScott.__name__, parameter=str(parameter))

    return locdata


def make_Matern(
    parent_intensity: int | float = 1,
    region: Region | npt.ArrayLike = (0, 1.0),
    cluster_mu: int | float = 1,
    radius: float | Sequence[float] = 1.0,
    clip: bool = True,
    shuffle: bool = True,
    seed: RandomGeneratorSeed = None,
) -> tuple[
    npt.NDArray[np.float64], npt.NDArray[np.int64], npt.NDArray[np.float64], Region
]:
    """
    Generate clustered point data following a Matern cluster random point
    process.
    Parent positions are distributed according to a homogeneous Poisson
    process with `parent_intensity`
    within the boundaries given by `region` expanded by the maximum radius.
    Each parent position is then replaced by spots of size `radius` with
    Poisson distributed points inside.
    Offspring from parent events that are located outside the region are
    included.

    Parameters
    ----------
    parent_intensity
        The intensity (points per unit region measure) of the Poisson point
        process for parent events.
    region
        The region (or support) for all features.
        If array-like it must provide upper and lower bounds for each feature.
    cluster_mu
        The mean number of points of the Poisson point process for
        cluster(cluster_mu) events.
    radius
        The radius for the spots. If tuple, the number of elements must be
        larger than the expected number of parents.
    clip
        If True the result will be clipped to 'region'. If False the extended
        region will be kept.
    shuffle
        If True shuffle the samples.
    seed
        random number generation seed

    Returns
    -------
    tuple[npt.NDArray[np.float64], npt.NDArray[np.int64], npt.NDArray[np.float64], Region]
       The generated samples, labels, parent_samples
       of shape (n_samples, n_features) and region
    """
    rng = np.random.default_rng(seed)

    if isinstance(region, EmptyRegion):
        samples, labels, parent_samples, region = (
            np.array([]),
            np.array([]),
            np.array([]),
            EmptyRegion(),
        )
        return samples, labels, parent_samples, region
    elif not isinstance(region, Region):
        region = Region.from_intervals(region)

    expansion_distance = np.max(radius)
    expanded_region = expand_region(region, expansion_distance)

    parent_samples = make_Poisson(
        intensity=parent_intensity, region=expanded_region, seed=rng
    )

    # radius: if radius is given as list, it must be consistent with the n_parent_samples
    if hasattr(radius, "__len__"):
        if len(radius) < len(parent_samples):
            raise ValueError(
                f"Length of `radius` {len(radius)} is less than "
                f"the generated n_parent_samples {len(parent_samples)}."
            )
        else:
            radii = radius[: len(parent_samples)]  # type: ignore
    else:  # if isinstance(radius, float):
        radii = np.full(len(parent_samples), radius)

    # replace parents by offspring samples
    samples_ = []
    labels_ = []
    for i, (parent, radius_) in enumerate(zip(parent_samples, radii)):
        if region.dimension == 1:
            offspring_region: Region = Interval(-radius_, radius_)
            offspring_intensity = cluster_mu / offspring_region.region_measure
            offspring_samples = make_Poisson(
                intensity=offspring_intensity, region=offspring_region, seed=rng
            )
        elif region.dimension == 2:
            offspring_region = Ellipse((0, 0), 2 * radius_, 2 * radius_, 0)
            offspring_intensity = cluster_mu / offspring_region.region_measure
            offspring_samples = make_Poisson(
                intensity=offspring_intensity, region=offspring_region, seed=rng
            )
        elif region.dimension == 3:
            raise NotImplementedError
        else:
            raise ValueError("region dimension must be 1, 2, or 3.")
        if len(offspring_samples) != 0:
            offspring_samples = offspring_samples + parent
            samples_.append(offspring_samples)
        labels_ += [i] * len(offspring_samples)

    samples = np.array(list(chain(*samples_)))
    labels = np.array(labels_)

    if clip is True:
        if len(samples) != 0:
            inside_indices = region.contains(samples)
            samples = samples[inside_indices]
            labels = labels[inside_indices]
        region_ = region
    else:
        region_ = expanded_region

    if shuffle:
        shuffled_indices = rng.permutation(len(samples))
        samples = samples[shuffled_indices]
        labels = labels[shuffled_indices]

    if (
        len(samples) == 0
    ):  # this is to convert empty arrays into arrays with shape (n_samples, n_features).
        samples = np.array([])
        samples = samples[:, np.newaxis]

    return samples, labels, parent_samples, region_


def simulate_Matern(
    parent_intensity: int | float = 1,
    region: Region | npt.ArrayLike = (0, 1.0),
    cluster_mu: int | float = 1,
    radius: float | Sequence[float] = 1.0,
    clip: bool = True,
    shuffle: bool = True,
    seed: RandomGeneratorSeed = None,
) -> LocData:
    """
    Generate clustered point data following a Matern cluster random point
    process.
    Parent positions are distributed according to a homogeneous Poisson
    process with `parent_intensity`
    within the boundaries given by `region` expanded by the maximum radius.
    Each parent position is then replaced by spots of size `radius` with
    Poisson distributed points inside.
    Offspring from parent events that are located outside the region are
    included.

    Parameters
    ----------
    parent_intensity
        The intensity (points per unit region measure) of the Poisson point
        process for parent events.
    region
        The region (or support) for each feature.
        If array-like it must provide upper and lower bounds for each feature.
    cluster_mu
        The mean number of points of the Poisson point process for
        cluster(cluster_mu) events.
    radius
        The radius for the spots. If tuple, the number of elements must be
        larger than the expected number of parents.
    clip
        If True the result will be clipped to 'region'.
        If False the extended region will be kept.
    shuffle
        If True shuffle the samples.
    seed
        random number generation seed

    Returns
    -------
    LocData
        The generated samples.
    """
    parameter = locals()
    samples, labels, _, region = make_Matern(
        parent_intensity, region, cluster_mu, radius, clip, shuffle, seed
    )
    region_ = region if isinstance(region, Region) else Region.from_intervals(region)  # type: ignore
    assert region_.dimension is not None  # type narrowing # noqa: S101
    locdata = LocData.from_coordinates(coordinates=samples)
    locdata.dimension = region_.dimension
    locdata.region = region_
    locdata.dataframe = locdata.dataframe.assign(cluster_label=labels)

    # metadata
    locdata.meta.source = metadata_pb2.SIMULATION
    del locdata.meta.history[:]
    locdata.meta.history.add(name=make_Matern.__name__, parameter=str(parameter))

    return locdata


def make_Thomas(
    parent_intensity: int | float = 1,
    region: Region | npt.ArrayLike = (0, 1.0),
    expansion_factor: int | float = 6,
    cluster_mu: int | float | Sequence[float] = 1,
    cluster_std: float | Sequence[float] | Sequence[Sequence[float]] = 1.0,
    clip: bool = True,
    shuffle: bool = True,
    seed: RandomGeneratorSeed = None,
) -> tuple[
    npt.NDArray[np.float64], npt.NDArray[np.int64], npt.NDArray[np.float64], Region
]:
    """
    Generate clustered point data following a Thomas random point process.
    Parent positions are distributed according to a homogeneous Poisson
    process with `parent_intensity`
    within the boundaries given by `region` expanded by an expansion distance
    that equals
    expansion_factor * max(cluster_std).
    Each parent position is then replaced by n offspring points
    where n is Poisson-distributed with mean number `cluster_mu`
    and point coordinates are normal-distributed around the parent point with
    standard deviation `cluster_std`.
    Offspring from parent events that are located outside the region are
    included.

    Parameters
    ----------
    parent_intensity
        The intensity (points per unit region measure) of the Poisson point
        process for parent events.
    region
        The region (or support) for all features.
        If array-like it must provide upper and lower bounds for each feature.
    expansion_factor
        Factor by which the cluster_std is multiplied to set the region
        expansion distance.
    cluster_mu
        The mean number of points for normal-distributed offspring points.
    cluster_std
        The standard deviation for normal-distributed offspring points.
    clip
        If True the result will be clipped to 'region'.
        If False the extended region will be kept.
    shuffle
        If True shuffle the samples.
    seed
        random number generation seed

    Returns
    -------
    tuple[npt.NDArray[np.float64], npt.NDArray[np.int64], npt.NDArray[np.float64], Region]
       The generated samples, labels, parent_samples
       of shape (n_samples, n_features) and region
    """
    rng = np.random.default_rng(seed)

    if not isinstance(region, Region):
        region = Region.from_intervals(region)

    if (
        parent_intensity == 0
        or (np.size(cluster_mu) == 1 and cluster_mu == 0)
        or isinstance(region, EmptyRegion)
    ):
        samples, labels, parent_samples, region = (
            np.array([]),
            np.array([]),
            np.array([]),
            region,
        )
        if region.dimension and region.dimension > 0:
            samples = samples[:, np.newaxis]
        return samples, labels, parent_samples, region

    # expand region
    expansion_distance = expansion_factor * np.max(cluster_std)
    expanded_region = expand_region(region, expansion_distance)

    parent_samples = make_Poisson(
        intensity=parent_intensity, region=expanded_region, seed=rng
    )
    n_cluster = len(parent_samples)

    # check cluster_std consistent with n_centers or n_features
    if len(np.shape(cluster_std)) == 0:
        cluster_std_ = np.full(
            shape=(n_cluster, region.dimension), fill_value=cluster_std  # type: ignore
        )
    elif len(np.shape(cluster_std)) == 1:  # iterate over cluster_std for each feature
        if region.dimension == 1 or len(cluster_std) != region.dimension:  # type: ignore
            raise TypeError(
                f"The shape of cluster_std {np.shape(cluster_std)} is incompatible "
                f"with n_features {region.dimension}."
            )
        else:
            cluster_std_ = np.empty(shape=(n_cluster, region.dimension))
            for i, element in enumerate(cluster_std):  # type: ignore
                cluster_std_[:, i] = np.full((n_cluster,), element)
    elif len(np.shape(cluster_std)) == 2:  # iterate over cluster_std for each center
        if np.shape(cluster_std) < (n_cluster, region.dimension):
            raise TypeError(
                f"The shape of cluster_std {np.shape(cluster_std)} is incompatible with "
                f"n_cluster {n_cluster} or n_features {region.dimension}."
            )
        else:
            cluster_std_ = cluster_std  # type: ignore
    else:
        raise TypeError(
            f"The shape of cluster_std {np.shape(cluster_std)} is incompatible."
        )

    # replace parents by normal-distributed offspring samples
    try:
        n_offspring_list = rng.poisson(lam=cluster_mu[:n_cluster], size=n_cluster)  # type: ignore
    except ValueError as e:
        e.args += (f"Too few offspring events for n_cluster: {n_cluster}",)
        raise
    except (TypeError, IndexError):
        n_offspring_list = rng.poisson(lam=cluster_mu, size=n_cluster)
    samples_ = []
    labels_ = []
    for i, (parent, std, n_offspring) in enumerate(
        zip(parent_samples, cluster_std_, n_offspring_list)
    ):
        offspring_samples = rng.normal(
            loc=parent, scale=std, size=(n_offspring, region.dimension)  # type: ignore
        )
        samples_.append(offspring_samples)
        labels_ += [i] * len(offspring_samples)

    samples = np.concatenate(samples_) if len(samples_) != 0 else np.array([])
    labels = np.array(labels_)

    if clip is True:
        if len(samples) != 0:
            inside_indices = region.contains(samples)
            samples = samples[inside_indices]
            labels = labels[inside_indices]
        region_ = region
    else:
        region_ = expanded_region

    if shuffle:
        shuffled_indices = rng.permutation(len(samples))
        samples = samples[shuffled_indices]
        labels = labels[shuffled_indices]

    if (
        len(samples) == 0
    ):  # this is to convert empty arrays into arrays with shape (n_samples, n_features).
        samples = np.array([])
        samples = samples[:, np.newaxis]

    return samples, labels, parent_samples, region_


def simulate_Thomas(
    parent_intensity: int | float = 1,
    region: Region | npt.ArrayLike = (0, 1.0),
    expansion_factor: int | float = 6,
    cluster_mu: int | float | Sequence[float] = 1,
    cluster_std: float | Sequence[float] | Sequence[Sequence[float]] = 1.0,
    clip: bool = True,
    shuffle: bool = True,
    seed: RandomGeneratorSeed = None,
) -> LocData:
    """
    Generate clustered point data following a Thomas random point process.
    Parent positions are distributed according to a homogeneous Poisson
    process with `parent_intensity`
    within the boundaries given by `region` expanded by an expansion distance
    that equals
    expansion_factor * max(cluster_std).
    Each parent position is then replaced by n offspring points
    where n is Poisson-distributed with mean number `cluster_mu`
    and point coordinates are normal-distributed around the parent point with
    standard deviation `cluster_std`.
    Offspring from parent events that are located outside the region are
    included.

    Parameters
    ----------
    parent_intensity
        The intensity (points per unit region measure) of the Poisson point
        process for parent events.
    region
        The region (or support) for each feature.
        If array-like it must provide upper and lower bounds for each feature.
    expansion_factor
        Factor by which the cluster_std is multiplied to set the region
        expansion distance.
    cluster_mu
        The mean number of points for normal-distributed offspring points.
    cluster_std
        The standard deviation for normal-distributed offspring points.
    clip
        If True the result will be clipped to 'region'.
        If False the extended region will be kept.
    shuffle
        If True shuffle the samples.
    seed
        random number generation seed

    Returns
    -------
    LocData
        The generated samples.
    """
    parameter = locals()
    samples, labels, _, region = make_Thomas(
        parent_intensity,
        region,
        expansion_factor,
        cluster_mu,
        cluster_std,
        clip,
        shuffle,
        seed,
    )
    region_ = region if isinstance(region, Region) else Region.from_intervals(region)  # type: ignore
    assert region_.dimension is not None  # type narrowing # noqa: S101
    locdata = LocData.from_coordinates(coordinates=samples)
    locdata.dimension = region_.dimension
    locdata.region = region_
    locdata.dataframe = locdata.dataframe.assign(cluster_label=labels)

    # metadata
    locdata.meta.source = metadata_pb2.SIMULATION
    del locdata.meta.history[:]
    locdata.meta.history.add(name=make_Thomas.__name__, parameter=str(parameter))

    return locdata


def make_dstorm(
    parent_intensity: int | float = 1,
    region: Region | npt.ArrayLike = (0, 1.0),
    expansion_factor: int | float = 6,
    cluster_mu: int | float | Sequence[float] = 1,
    min_points: int = 0,
    cluster_std: float | Sequence[float] | Sequence[Sequence[float]] = 1.0,
    clip: bool = True,
    shuffle: bool = True,
    seed: RandomGeneratorSeed = None,
) -> tuple[
    npt.NDArray[np.float64], npt.NDArray[np.int64], npt.NDArray[np.float64], Region
]:
    """
    Generate clustered point data following a Thomas-like random point process.
    Parent positions are distributed according to a homogeneous Poisson
    process with `parent_intensity`
    within the boundaries given by `region` expanded by an expansion distance
    that equals
    expansion_factor * max(cluster_std).
    Each parent position is then replaced by n offspring points
    where n is geometrically-distributed with mean number `cluster_mu`
    and point coordinates are normal-distributed around the parent point with
    standard deviation `cluster_std`.
    Offspring from parent events that are located outside the region are
    included.

    Parameters
    ----------
    parent_intensity
        The intensity (points per unit region measure) of the Poisson point
        process for parent events.
    region
        The region (or support) for all features.
        If array-like it must provide upper and lower bounds for each feature.
    expansion_factor
        Factor by which the cluster_std is multiplied to set the region
        expansion distance.
    cluster_mu
        The mean number of points for normal-distributed offspring points.
    min_points
        The minimum number of points per cluster.
    cluster_std
        The standard deviation for normal-distributed offspring points.
    clip
        If True the result will be clipped to 'region'. If False the extended
        region will be kept.
    shuffle
        If True shuffle the samples.
    seed
        random number generation seed

    Returns
    -------
    tuple[npt.NDArray[np.float64], npt.NDArray[np.int64], npt.NDArray[np.float64], Region]
       The generated samples, labels, parent_samples
        of shape (n_samples, n_features) and region
    """
    rng = np.random.default_rng(seed)

    if not isinstance(region, Region):
        region = Region.from_intervals(region)

    if (
        parent_intensity == 0
        or (np.size(cluster_mu) == 1 and cluster_mu == 0)
        or isinstance(region, EmptyRegion)
    ):
        samples, labels, parent_samples, region = (
            np.array([]),
            np.array([]),
            np.array([]),
            region,
        )
        if region.dimension and region.dimension > 0:
            samples = samples[:, np.newaxis]
        return samples, labels, parent_samples, region

    # expand region
    expansion_distance = expansion_factor * np.max(cluster_std)
    expanded_region = expand_region(region, expansion_distance)

    parent_samples = make_Poisson(
        intensity=parent_intensity, region=expanded_region, seed=rng
    )
    n_cluster = len(parent_samples)

    # check cluster_std consistent with n_centers or n_features
    if len(np.shape(cluster_std)) == 0:
        cluster_std_ = np.full(
            shape=(n_cluster, region.dimension), fill_value=cluster_std  # type: ignore
        )
    elif len(np.shape(cluster_std)) == 1:  # iterate over cluster_std for each feature
        if region.dimension == 1 or len(cluster_std) != region.dimension:  # type: ignore
            raise TypeError(
                f"The shape of cluster_std {np.shape(cluster_std)} is incompatible "
                f"with n_features {region.dimension}."
            )
        else:
            cluster_std_ = np.empty(shape=(n_cluster, region.dimension))
            for i, element in enumerate(cluster_std):  # type: ignore
                cluster_std_[:, i] = np.full((n_cluster,), element)
    elif len(np.shape(cluster_std)) == 2:  # iterate over cluster_std for each center
        if np.shape(cluster_std) < (n_cluster, region.dimension):
            raise TypeError(
                f"The shape of cluster_std {np.shape(cluster_std)} is incompatible with "
                f"n_cluster {n_cluster} or n_features {region.dimension}."
            )
        else:
            cluster_std_ = cluster_std  # type: ignore
    else:
        raise TypeError(
            f"The shape of cluster_std {np.shape(cluster_std)} is incompatible."
        )

    # replace parents by normal-distributed offspring samples
    try:
        p_values = [1 / (mu + 1 - min_points) for mu in cluster_mu[:n_cluster]]  # type: ignore
        # p for a geometric distribution sampling points from min_points to inf is 1 / (mean + 1 - min_points)
        n_offspring_list = rng.geometric(p=p_values, size=n_cluster) - 1 + min_points
        # rng.geometric samples points from 1 to inf.
    except ValueError as e:
        e.args += (f"Too few offspring events for n_cluster: {n_cluster}",)
        raise
    except (TypeError, IndexError):
        n_offspring_list = (
            rng.geometric(p=1 / (cluster_mu + 1 - min_points), size=n_cluster)  # type: ignore
            - 1
            + min_points
        )

    samples_ = []
    labels_ = []
    for i, (parent, std, n_offspring) in enumerate(
        zip(parent_samples, cluster_std_, n_offspring_list)
    ):
        offspring_samples = rng.normal(
            loc=parent, scale=std, size=(n_offspring, region.dimension)  # type: ignore
        )
        samples_.append(offspring_samples)
        labels_ += [i] * len(offspring_samples)

    samples = np.concatenate(samples_) if len(samples_) != 0 else np.array([])
    labels = np.array(labels_)

    if clip is True:
        if len(samples) != 0:
            inside_indices = region.contains(samples)
            samples = samples[inside_indices]
            labels = labels[inside_indices]
        region_ = region
    else:
        region_ = expanded_region

    if shuffle:
        shuffled_indices = rng.permutation(len(samples))
        samples = samples[shuffled_indices]
        labels = labels[shuffled_indices]

    if (
        len(samples) == 0
    ):  # this is to convert empty arrays into arrays with shape (n_samples, n_features).
        samples = np.array([])
        samples = samples[:, np.newaxis]

    return samples, labels, parent_samples, region_


def simulate_dstorm(
    parent_intensity: int | float = 1,
    region: Region | npt.ArrayLike = (0, 1.0),
    expansion_factor: int | float = 6,
    cluster_mu: int | float | Sequence[float] = 1,
    min_points: int = 0,
    cluster_std: float | Sequence[float] | Sequence[Sequence[float]] = 1.0,
    clip: bool = True,
    shuffle: bool = True,
    seed: RandomGeneratorSeed = None,
) -> LocData:
    """
    Generate clustered point data following a Thomas-like random point process.
    Parent positions are distributed according to a homogeneous Poisson
    process with `parent_intensity`
    within the boundaries given by `region` expanded by an expansion distance
    that equals
    expansion_factor * max(cluster_std).
    Each parent position is then replaced by n offspring points
    where n is geometrically-distributed with mean number `cluster_mu`
    and point coordinates are normal-distributed around the parent point with
    standard deviation `cluster_std`.
    Offspring from parent events that are located outside the region are
    included.

    Parameters
    ----------
    parent_intensity
        The intensity (points per unit region measure) of the Poisson point
        process for parent events.
    region
        The region (or support) for each feature.
        If array-like it must provide upper and lower bounds for each feature.
    expansion_factor
        Factor by which the cluster_std is multiplied to set the region
        expansion distance.
    cluster_mu
        The mean number of points for normal-distributed offspring points.
    min_points
        The minimum number of points per cluster.
    cluster_std
        The standard deviation for normal-distributed offspring points.
    clip
        If True the result will be clipped to 'region'.
        If False the extended region will be kept.
    shuffle
        If True shuffle the samples.
    seed
        random number generation seed

    Returns
    -------
    LocData
        The generated samples.
    """
    parameter = locals()
    samples, labels, _, region = make_dstorm(
        parent_intensity,
        region,
        expansion_factor,
        cluster_mu,
        min_points,
        cluster_std,
        clip,
        shuffle,
        seed,
    )
    region_ = region if isinstance(region, Region) else Region.from_intervals(region)  # type: ignore
    assert region_.dimension is not None  # type narrowing # noqa: S101
    locdata = LocData.from_coordinates(coordinates=samples)
    locdata.dimension = region_.dimension
    locdata.region = region_
    locdata.dataframe = locdata.dataframe.assign(cluster_label=labels)

    # metadata
    locdata.meta.source = metadata_pb2.SIMULATION
    del locdata.meta.history[:]
    locdata.meta.history.add(name=make_dstorm.__name__, parameter=str(parameter))

    return locdata


def _random_walk(
    n_walks: int = 1,
    n_steps: int = 10,
    dimensions: int = 2,
    diffusion_constant: int | float = 1,
    time_step: float = 10,
    seed: RandomGeneratorSeed = None,
) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
    """
    Random walk simulation

    Parameters
    ----------
    n_walks
        Number of walks
    n_steps
        Number of time steps (i.e. frames)
    dimensions
        spatial dimensions to simulate
    diffusion_constant
        Diffusion constant in units length per seconds^2
    time_step
        Time per frame (or simulation step) in seconds.
    seed
        random number generation seed

    Returns
    -------
    tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]
        (times, positions), where shape(times) is 1 and shape of positions
        is (n_walks, n_steps, dimensions)
    """
    rng = np.random.default_rng(seed)
    # equally spaced time steps
    times = np.arange(n_steps, dtype=float) * time_step

    # random step sizes according to the diffusion constant
    random_numbers = rng.integers(
        0, 2, size=(n_walks, n_steps, dimensions)
    )  # np.random.randint(0, 2, size=(n_walks, n_steps, dimensions))
    step_size = np.sqrt(2 * dimensions * diffusion_constant * time_step)
    steps = np.where(random_numbers == 0, -step_size, +step_size)

    # walker positions
    positions = np.cumsum(steps, axis=1)

    return times, positions


def simulate_tracks(
    n_walks: int = 1,
    n_steps: int = 10,
    ranges: tuple[tuple[int, int], tuple[int, int]] = ((0, 10000), (0, 10000)),
    diffusion_constant: int | float = 1,
    time_step: float = 10,
    seed: RandomGeneratorSeed = None,
) -> LocData:
    """
    Provide a dataset of localizations representing random walks with
    starting points being spatially-distributed
    on a rectangular shape or cubic volume by complete spatial randomness.

    Parameters
    ----------
    n_walks
        Number of walks
    n_steps
        Number of time steps (i.e. frames)
    ranges
        the range for valid x[, y, z]-coordinates
    diffusion_constant
        Diffusion constant with unit length per seconds^2
    time_step
        Time per frame (or simulation step) in seconds.
    seed
        random number generation seed

    Returns
    -------
    LocData
        A new LocData instance with localization data.
    """
    parameter = locals()

    rng = np.random.default_rng(seed)

    start_positions = np.array(
        [rng.uniform(*_range, size=n_walks) for _range in ranges]
    ).T

    times, positions = _random_walk(
        n_walks=n_walks,
        n_steps=n_steps,
        dimensions=len(ranges),
        diffusion_constant=diffusion_constant,
        time_step=time_step,
        seed=rng,
    )

    new_positions = np.concatenate(
        [
            start_position + position
            for start_position, position in zip(start_positions, positions)
        ]
    )

    locdata_dict = {
        "position_" + label: position_values
        for _, position_values, label in zip(ranges, new_positions.T, ("x", "y", "z"))
    }

    locdata_dict.update(frame=np.tile(range(len(times)), n_walks))

    locdata = LocData.from_dataframe(dataframe=pd.DataFrame(locdata_dict))

    # metadata
    locdata.meta.source = metadata_pb2.SIMULATION
    del locdata.meta.history[:]
    locdata.meta.history.add(
        name=sys._getframe().f_code.co_name, parameter=str(parameter)
    )

    return locdata


def resample(
    locdata: LocData, n_samples: int = 10, seed: RandomGeneratorSeed = None
) -> LocData:
    """
    Resample locdata according to localization uncertainty.
    Per localization `n_samples` new localizations
    are simulated normally distributed around the localization coordinates
    with a standard deviation set to the uncertainty in each dimension.
    Uncertainties are taken from "uncertainty_c" or "uncertainty".
    The resulting LocData object carries new localizations with the following
    new properties: position coordinates, 'original_index'.

    Parameters
    ----------
    locdata
        Localization data to be resampled
    n_samples
        The number of localizations generated for each original localization.
    seed
        random number generation seed

    Returns
    -------
    LocData
        New localization data with simulated coordinates.
    """
    rng = np.random.default_rng(seed)

    available_coordinate_keys = _get_loc_property_key_per_dimension(
        locdata=locdata.data, property_key="position"
    )
    available_uncertainty_keys = _get_loc_property_key_per_dimension(
        locdata=locdata.data, property_key="uncertainty"
    )

    original_index_label = _bump_property_key(
        loc_property="original_index", loc_properties=locdata.data.columns
    )
    new_df = locdata.data.loc[locdata.data.index.repeat(n_samples)].reset_index(  # type: ignore
        names=original_index_label
    )

    for c_label, u_label in zip(available_coordinate_keys, available_uncertainty_keys):
        if c_label is not None:
            if u_label is None:
                logger.warning(f"No uncertainties available for {c_label}.")
            else:
                new_df[c_label] = rng.normal(loc=new_df[c_label], scale=new_df[u_label])

    # metadata
    meta_ = metadata_pb2.Metadata()
    meta_.CopyFrom(locdata.meta)
    try:
        meta_.ClearField("identifier")
    except ValueError:
        pass

    try:
        meta_.ClearField("element_count")
    except ValueError:
        pass

    try:
        meta_.ClearField("frame_count")
    except ValueError:
        pass

    meta_.modification_time.GetCurrentTime()
    meta_.state = metadata_pb2.MODIFIED
    meta_.ancestor_identifiers.append(locdata.meta.identifier)
    meta_.history.add(
        name="resample", parameter=f"locdata={locdata}, n_samples={n_samples}"
    )

    # instantiate
    new_locdata = LocData.from_dataframe(dataframe=new_df, meta=meta_)

    return new_locdata


def _random_poisson_repetitions(
    n_samples: int, lam: float, seed: RandomGeneratorSeed = None
) -> npt.NDArray[np.int64]:
    """
    Return numpy.ndarray of sorted integers with each integer i being
    repeated n(i) times
    where n(i) is drawn from a Poisson distribution with mean `lam`.

    Parameters
    ----------
    n_samples
        number of elements to be returned
    lam
        mean of the Poisson distribution (lambda)
    seed
        random number generation seed

    Returns
    -------
    npt.NDArray[np.int64]
        The generated sequence of integers with shape (n_samples,)
    """
    rng = np.random.default_rng(seed)

    frames = np.ones(n_samples, dtype=int)
    n_random_numbers = n_samples if lam > 0 else int(n_samples / lam)
    position = 0
    current_number = 0
    while position < n_samples:
        repeats = rng.poisson(lam=lam, size=n_random_numbers)
        for repeat in repeats:
            try:
                frames[position : position + repeat] = current_number
            except ValueError:
                frames[position:] = current_number
                break
            position += repeat
            current_number += 1
    return frames


def simulate_frame_numbers(
    n_samples: int, lam: float, seed: RandomGeneratorSeed = None
) -> npt.NDArray[np.int64]:
    """
    Simulate Poisson-distributed frame numbers for a list of localizations.

    Return numpy.ndarray of sorted integers with each integer i being
    repeated n(i) times
    where n(i) is drawn from a Poisson distribution with mean `lam`.

    Use the following to add frame numbers to a given LocData object::

        frames = simulate_frame_numbers(n_samples=len(locdata), lam=5)
        locdata.dataframe = locdata.dataframe.assign(frame = frames)

    Parameters
    ----------
    n_samples
        number of elements to be returned
    lam
        mean of the Poisson distribution (lambda)
    seed
        random number generation seed

    Returns
    -------
    npt.NDArray[np.int64]
        The generated sequence of integers with shape (n_samples,)
    """
    return _random_poisson_repetitions(n_samples, lam, seed=seed)


def randomize(
    locdata: LocData,
    hull_region: Region | Literal["bb", "ch", "as", "obb"] = "bb",
    seed: RandomGeneratorSeed = None,
) -> LocData:
    """
    Transform locdata coordinates into randomized coordinates that follow
    complete spatial randomness on the same region as the input locdata.

    Parameters
    ----------
    locdata
        Localization data to be randomized
    hull_region
        Region of interest.
        String identifier can refer to the corresponding hull.
    seed
        random number generation seed

    Returns
    -------
    LocData
        New localization data with randomized coordinates.
    """
    # todo: fix treatment of empty locdata
    local_parameter = locals()

    rng = np.random.default_rng(seed)

    try:
        if hull_region == "bb":
            region_ = locdata.bounding_box.hull.T  # type: ignore[union-attr]
        elif hull_region == "ch":
            region_ = locdata.convex_hull.region  # type: ignore[union-attr]
        elif hull_region == "as":
            region_ = locdata.alpha_shape.region  # type: ignore[union-attr]
        elif hull_region == "obb":
            region_ = locdata.oriented_bounding_box.region  # type: ignore[union-attr]
        elif isinstance(hull_region, Region):
            region_ = hull_region
        else:
            raise ValueError

        new_locdata = simulate_uniform(n_samples=len(locdata), region=region_, seed=rng)

    except (AttributeError, ValueError, TypeError) as exception:
        raise AttributeError(f"Region {hull_region} is not available.") from exception

    # update metadata
    meta_ = _modify_meta(
        locdata,
        new_locdata,
        function_name=sys._getframe().f_code.co_name,
        parameter=local_parameter,
        meta=None,
    )
    new_locdata.meta = meta_

    return new_locdata



"""

This module provides functions to interact with napari
for rendering locdata objects in 2D.

"""

from __future__ import annotations

import logging
from collections.abc import Callable, Iterable, Sequence
from typing import TYPE_CHECKING, Any, Literal

import numpy as np
from matplotlib import colors as mcolors

from locan import locdata_id
from locan.configuration import COLORMAP_DEFAULTS
from locan.data import LocData
from locan.data.locdata_utils import _check_loc_properties
from locan.dependencies import HAS_DEPENDENCY, needs_package
from locan.process.aggregate import Bins, histogram
from locan.visualize.colormap import ColormapType, get_colormap
from locan.visualize.transform import Trafo, adjust_contrast

if HAS_DEPENDENCY["napari"]:
    import napari

if TYPE_CHECKING:
    import boost_histogram as bh

__all__: list[str] = [
    "render_2d_napari",
    "render_2d_napari_image",
    "render_2d_rgb_napari",
]

logger = logging.getLogger(__name__)


@needs_package("napari")
def render_2d_napari_image(
    locdata: LocData,
    loc_properties: list[str] | None = None,
    other_property: str | None = None,
    bins: Bins | bh.axis.Axis | bh.axis.AxesTuple | None = None,
    n_bins: int | Sequence[int] | None = None,
    bin_size: float | Sequence[float] | Sequence[Sequence[float]] | None = None,
    bin_edges: Sequence[float] | Sequence[Sequence[float]] | None = None,
    bin_range: (
        tuple[float, float]
        | Sequence[float]
        | Sequence[Sequence[float]]
        | Literal["zero", "link"]
        | None
    ) = None,
    rescale: int | str | Trafo | Callable[..., Any] | bool | None = None,
    cmap: ColormapType = COLORMAP_DEFAULTS["CONTINUOUS"],
    **kwargs: Any,
) -> napari.types.LayerData:
    """
    Render localization data into a 2D image by binning x,y-coordinates into
    regular bins. Provide layer data for napari.

    Parameters
    ----------
    locdata
        Localization data.
    loc_properties
        Localization properties to be grouped into bins.
        If None The coordinate_values of `locdata` are used.
    other_property
        Localization property (columns in `locdata.data`) that is averaged
        in each pixel.
        If None, localization counts are shown.
    bins
        The bin specification as defined in :class:`Bins`
    bin_edges
        Bin edges for all or each dimension
        with shape (dimension, n_bin_edges).
    bin_range
        Minimum and maximum edge for all or each dimensions
        with shape (2,) or (dimension, 2).
        If None (min, max) ranges are determined from data and returned;
        if 'zero' (0, max) ranges with max determined from data are returned.
        if 'link' (min_all, max_all) ranges with min and max determined from
        all combined data are returned.
    n_bins
        The number of bins for all or each dimension.
        5 yields 5 bins in all dimensions.
        (2, 5) yields 2 bins for one dimension and 5 for the other dimension.
    bin_size
        The size of bins for all or each bin and for all or each dimension
        with shape (dimension,) or (dimension, n_bins).
        5 would describe bin_size of 5 for all bins in all dimensions.
        ((2, 5),) yield bins of size (2, 5) for one dimension.
        (2, 5) yields bins of size 2 for one dimension and 5 for the other
        dimension.
        ((2, 5), (1, 3)) yields bins of size (2, 5) for one dimension and
        (1, 3) for the other dimension.
        To specify arbitrary sequence of `bin_size` use `bin_edges` instead.
    rescale
        Transformation as defined in :class:`locan.Trafo` or by
        transformation function.
        For None or False no rescaling occurs.
        Legacy behavior:
        For tuple with upper and lower bounds provided in percent,
        rescale intensity values to be within percentile of max and min
        intensities.
        For 'equal' intensity values are rescaled by histogram equalization.
    cmap
        The Colormap object used to map normalized data values to RGBA colors.
    kwargs
        Other parameters passed to :func:`napari.Viewer.add_image`.

    Returns
    -------
    napari.types.LayerData
        Tuple with data, image_kwargs, layer_type="image"
    """
    # raise if no or single point in locdata
    if len(locdata) < 2:
        if len(locdata) == 1:
            logger.warning("Locdata carries a single localization.")
        raise ValueError(
            "Locdata has zero or one localizations - must have more than one."
        )

    data, bins, _labels = histogram(
        locdata=locdata,
        loc_properties=loc_properties,
        other_property=other_property,
        bins=bins,
        n_bins=n_bins,
        bin_size=bin_size,
        bin_edges=bin_edges,
        bin_range=bin_range,
    )
    if not all(bins.is_equally_sized):
        raise ValueError("All bins must be equally sized.")
    data = adjust_contrast(data, rescale)

    add_image_kwargs = dict(
        name=f"LocData {locdata_id}",
        colormap=get_colormap(colormap=cmap).napari,
        scale=bins.bin_size,
        translate=np.asarray(bins.bin_range)[:, 0] + np.asarray(bins.bin_size) / 2,
        metadata=dict(message=locdata.meta.SerializeToString()),
    )

    layer_data = (data, dict(add_image_kwargs, **kwargs), "image")
    return layer_data


@needs_package("napari")
def render_2d_napari(
    locdata: LocData,
    loc_properties: list[str] | None = None,
    other_property: str | None = None,
    bins: Bins | bh.axis.Axis | bh.axis.AxesTuple | None = None,
    n_bins: int | Sequence[int] | None = None,
    bin_size: float | Sequence[float] | Sequence[Sequence[float]] | None = None,
    bin_edges: Sequence[float] | Sequence[Sequence[float]] | None = None,
    bin_range: (
        tuple[float, float]
        | Sequence[float]
        | Sequence[Sequence[float]]
        | Literal["zero", "link"]
        | None
    ) = None,
    rescale: int | str | Trafo | Callable[..., Any] | bool | None = None,
    viewer: napari.Viewer = None,
    cmap: ColormapType = COLORMAP_DEFAULTS["CONTINUOUS"],
    **kwargs: Any,
) -> napari.Viewer:
    """
    Render localization data into a 2D image by binning x,y-coordinates into
    regular bins. Render the data using napari.

    Parameters
    ----------
    locdata
        Localization data.
    loc_properties
        Localization properties to be grouped into bins.
        If None The coordinate_values of `locdata` are used.
    other_property
        Localization property (columns in `locdata.data`) that is averaged
        in each pixel.
        If None, localization counts are shown.
    bins
        The bin specification as defined in :class:`Bins`
    bin_edges
        Bin edges for all or each dimension
        with shape (dimension, n_bin_edges).
    bin_range
        Minimum and maximum edge for all or each dimensions
        with shape (2,) or (dimension, 2).
        If None (min, max) ranges are determined from data and returned;
        if 'zero' (0, max) ranges with max determined from data are returned.
        if 'link' (min_all, max_all) ranges with min and max determined from
        all combined data are returned.
    n_bins
        The number of bins for all or each dimension.
        5 yields 5 bins in all dimensions.
        (2, 5) yields 2 bins for one dimension and 5 for the other dimension.
    bin_size
        The size of bins for all or each bin and for all or each dimension
        with shape (dimension,) or (dimension, n_bins).
        5 would describe bin_size of 5 for all bins in all dimensions.
        ((2, 5),) yield bins of size (2, 5) for one dimension.
        (2, 5) yields bins of size 2 for one dimension and 5 for the other
        dimension.
        ((2, 5), (1, 3)) yields bins of size (2, 5) for one dimension and
        (1, 3) for the other dimension.
        To specify arbitrary sequence of `bin_size` use `bin_edges` instead.
    rescale
        Transformation as defined in :class:`locan.Trafo` or by
        transformation function.
        For None or False no rescaling occurs.
        Legacy behavior:
        For tuple with upper and lower bounds provided in percent,
        rescale intensity values to be within percentile of max and min
        intensities.
        For 'equal' intensity values are rescaled by histogram equalization.
    viewer
        The viewer object on which to add the image
    cmap : str | Colormap
        The Colormap object used to map normalized data values to RGBA colors.
    kwargs
        Other parameters passed to :func:`napari.Viewer.add_image`.

    Returns
    -------
    napari.Viewer
    """
    # Provide napari viewer if not provided
    if viewer is None:
        viewer = napari.Viewer()

    try:
        data, image_kwargs, _layer_type = render_2d_napari_image(
            locdata=locdata,
            loc_properties=loc_properties,
            other_property=other_property,
            bins=bins,
            n_bins=n_bins,
            bin_size=bin_size,
            bin_edges=bin_edges,
            bin_range=bin_range,
            rescale=rescale,
            cmap=get_colormap(colormap=cmap).napari,
            **kwargs,
        )
        viewer.add_image(data=data, **dict(image_kwargs, **kwargs))

        # set scale_bar unit
        names = list(
            set(
                item.unit
                for item in locdata.meta.localization_properties
                if item.name in locdata.coordinate_keys
            )
        )
        if len(names) == 1:
            viewer.scale_bar.unit = f"1 {names[0]}"
        else:
            viewer.scale_bar.unit = None
    except ValueError as e:
        if (
            len(e.args) > 0
            and e.args[0]
            == "Locdata has zero or one localizations - must have more than one."
        ):
            pass
        else:
            raise e
    return viewer


@needs_package("napari")
def render_2d_rgb_napari(
    locdatas: Iterable[LocData],
    loc_properties: list[str] | None = None,
    other_property: str | None = None,
    bins: Bins | bh.axis.Axis | bh.axis.AxesTuple | None = None,
    n_bins: int | Sequence[int] | None = None,
    bin_size: float | Sequence[float] | Sequence[Sequence[float]] | None = 10,
    bin_edges: Sequence[float] | Sequence[Sequence[float]] | None = None,
    bin_range: (
        tuple[float, float]
        | Sequence[float]
        | Sequence[Sequence[float]]
        | Literal["zero", "link"]
        | None
    ) = None,
    rescale: int | str | Trafo | Callable[..., Any] | bool | None = None,
    viewer: napari.Viewer = None,
    **kwargs: Any,
) -> napari.Viewer:
    """
    Render localization data into a 2D RGB image by binning x,y-coordinates
    into regular bins.

    Note
    ----
    For rescale=False no normalization is carried out image intensities are
    clipped to (0, 1) for float value or (0, 255) for integer values according
    to the matplotlib.imshow behavior.
    For rescale=None we apply a normalization to (min, max) of all intensity
    values.
    For all other rescale options the normalization is applied to each
    individual image.

    Parameters
    ----------
    locdatas
        Localization data.
    loc_properties
        Localization properties to be grouped into bins. If None
        The coordinate_values of `locdata` are used.
    other_property
        Localization property (columns in `locdata.data`) that is averaged in
        each pixel.
        If None, localization counts are shown.
    bins
        The bin specification as defined in :class:`Bins`
    bin_edges
        Bin edges for all or each dimension
        with shape (dimension, n_bin_edges).
    bin_range
        Minimum and maximum edge for all or each dimensions
        with shape (2,) or (dimension, 2).
        If None (min, max) ranges are determined from data and returned;
        if 'zero' (0, max) ranges with max determined from data are returned.
        if 'link' (min_all, max_all) ranges with min and max determined from
        all combined data are returned.
    n_bins
        The number of bins for all or each dimension.
        5 yields 5 bins in all dimensions.
        (2, 5) yields 2 bins for one dimension and 5 for the other dimension.
    bin_size
        The size of bins for all or each bin and for all or each dimension
        with shape (dimension,) or (dimension, n_bins).
        5 would describe bin_size of 5 for all bins in all dimensions.
        ((2, 5),) yield bins of size (2, 5) for one dimension.
        (2, 5) yields bins of size 2 for one dimension and 5 for the other
        dimension.
        ((2, 5), (1, 3)) yields bins of size (2, 5) for one dimension and
        (1, 3) for the other dimension.
        To specify arbitrary sequence of `bin_size` use `bin_edges` instead.
    rescale
        Transformation as defined in :class:`locan.Trafo` or by
        transformation function.
        For None or False no rescaling occurs.
        Legacy behavior:
        For tuple with upper and lower bounds provided in percent,
        rescale intensity values to be within percentile of max and min
        intensities.
        For 'equal' intensity values are rescaled by histogram equalization.
    viewer
        The viewer object on which to add the image
    kwargs
        Other parameters passed to :func:`napari.Viewer.add_image`.

    Returns
    -------
    napari.Viewer
    """
    # Provide napari viewer if not provided
    if viewer is None:
        viewer = napari.Viewer()

    locdata_temp = LocData.concat(locdatas)

    # return viewer if no or single point in locdata
    if len(locdata_temp) < 2:
        if len(locdata_temp) == 1:
            logger.warning("Locdata carries a single localization.")
        return viewer

    if bin_edges is None:
        _, bins, labels = histogram(
            locdata=locdata_temp,
            loc_properties=loc_properties,
            other_property=other_property,
            bins=bins,
            n_bins=n_bins,
            bin_size=bin_size,
            bin_edges=bin_edges,
            bin_range=bin_range,
        )

    else:
        labels = _check_loc_properties(locdata_temp, loc_properties)
        bins = Bins(bin_edges=bin_edges, labels=labels)

    imgs = [
        histogram(
            locdata=locdata,
            loc_properties=loc_properties,
            other_property=other_property,
            bin_edges=bins.bin_edges,  # type: ignore
        ).data
        for locdata in locdatas
    ]

    if rescale is None:
        norm: int | str | Trafo | Callable[..., Any] = mcolors.Normalize(
            vmin=np.min(imgs), vmax=np.max(imgs)
        )
    else:
        norm = rescale
    imgs = [adjust_contrast(img, rescale=norm) for img in imgs]

    new = np.zeros_like(imgs[0])
    rgb_stack = np.stack([new] * 3, axis=2)

    for i, img in enumerate(imgs):
        rgb_stack[:, :, i] = img

    rgb_stack = np.transpose(rgb_stack, axes=(1, 0, 2))
    viewer.add_image(
        rgb_stack,
        **dict({"name": f"LocData {locdata_id}", "rgb": True}, **kwargs),
    )
    return viewer



import numpy as np
import pandas as pd
import pytest

from locan import LocData
from locan.analysis import LocalizationUncertainty
from locan.analysis.uncertainty import (
    _localization_uncertainty,
    localization_precision_model_1,
    localization_precision_model_2,
    localization_precision_model_3,
)


@pytest.fixture()
def locdata_simple():
    locdata_dict = {
        "position_x": [0, 0, 1, 4, 5],
        "position_y": [0, 1, 3, 4, 1],
        "intensity": [0, 1**2, 3**2, 4**2, 1],
        "psf_sigma_x": [100, 100, 100, 100, 100],
        "pixel_size": [10, 10, 10, 10, 10],
        "local_background": [0.1, 0.1, 0.1, 0.1, 0.1],
    }
    return LocData(dataframe=pd.DataFrame.from_dict(locdata_dict))


def test_localization_precision_model_1():
    intensities = [1**2, 2**2, 3**2]
    uncertainties = localization_precision_model_1(intensity=intensities)
    assert isinstance(uncertainties, np.ndarray)
    assert np.array_equal(uncertainties, [1, 0.5, 1 / 3])


def test_localization_precision_model_2():
    intensities = [1**2, 2**2, 3**2]
    uncertainties = localization_precision_model_2(intensity=intensities, psf_sigma=12)
    assert isinstance(uncertainties, np.ndarray)
    assert np.array_equal(uncertainties, [12, 6.0, 4])

    sigma_psf_list = [12, 12, 12]
    uncertainties = localization_precision_model_2(
        intensity=intensities, psf_sigma=sigma_psf_list
    )
    assert isinstance(uncertainties, np.ndarray)
    assert np.array_equal(uncertainties, [12, 6.0, 4])


def test_localization_precision_model_3():
    intensities = [1**2, 2**2, 3**2]
    uncertainties = localization_precision_model_3(
        intensity=intensities, psf_sigma=12, pixel_size=10, local_background=0.1
    )
    assert isinstance(uncertainties, np.ndarray)
    np.testing.assert_allclose(uncertainties, [28.8, 9.6, 5.5], rtol=1e-1)

    sigma_psf_list = [12, 12, 12]
    background_list = [0.1, 0.1, 0.1]
    uncertainties = localization_precision_model_3(
        intensity=intensities,
        psf_sigma=sigma_psf_list,
        pixel_size=10,
        local_background=background_list,
    )
    assert isinstance(uncertainties, np.ndarray)
    np.testing.assert_allclose(uncertainties, [28.8, 9.6, 5.5], rtol=1e-1)

    uncertainties = localization_precision_model_3(
        intensity=2000,
        psf_sigma=150,
        pixel_size=130,
        local_background=100,
    )
    assert uncertainties == pytest.approx(6.32, rel=0.01)


def test__localization_uncertainty(locdata_simple, caplog):
    results = _localization_uncertainty(locdata=locdata_simple, model=1)
    assert caplog.record_tuples == [
        (
            "locan.analysis.uncertainty",
            30,
            "The localization property `intensity` does not have the unit photons.",
        )
    ]
    assert all(key_ in results.columns for key_ in ["uncertainty_x", "uncertainty_y"])

    with pytest.warns(RuntimeWarning):
        results = _localization_uncertainty(locdata=locdata_simple, model=2)
    assert all(key_ in results.columns for key_ in ["uncertainty_x"])
    assert results.iloc[-1, 0] == pytest.approx(100, rel=0.01)

    with pytest.warns(RuntimeWarning):
        results = _localization_uncertainty(
            locdata=locdata_simple, model=2, psf_sigma_x=150
        )
    assert all(key_ in results.columns for key_ in ["uncertainty_x"])
    assert results.iloc[-1, 0] == pytest.approx(150, rel=0.01)

    with pytest.warns(RuntimeWarning):
        results = _localization_uncertainty(locdata=locdata_simple, model=3)
    assert all(key_ in results.columns for key_ in ["uncertainty_x"])
    assert results.iloc[-1, 0] == pytest.approx(1592.0, rel=0.01)

    with pytest.warns(RuntimeWarning):
        results = _localization_uncertainty(
            locdata=locdata_simple, model=3, pixel_size=130
        )
    assert all(key_ in results.columns for key_ in ["uncertainty_x"])
    assert results.iloc[-1, 0] == pytest.approx(192.8, rel=0.01)

    def my_callable(intensity, factor):
        intensity = np.asarray(intensity)
        return intensity * factor

    results = _localization_uncertainty(
        locdata=locdata_simple, model=my_callable, factor=2
    )
    assert all(key_ in results.columns for key_ in ["uncertainty_x", "uncertainty_y"])
    assert results.iloc[-1, 0] == pytest.approx(2, rel=0.01)

    with pytest.raises(KeyError):
        _localization_uncertainty(
            locdata=locdata_simple, model=my_callable, not_existing=2
        )

    results = _localization_uncertainty(locdata=LocData(), model=1)
    assert results.empty


class TestLocalizationUncertainty:
    def test_empty_locdata(self, caplog):
        LocalizationUncertainty().compute(locdata=LocData())
        assert caplog.record_tuples == [
            ("locan.analysis.uncertainty", 30, "Locdata is empty.")
        ]

    def test_init(self, locdata_simple):
        with pytest.warns(RuntimeWarning):
            unc = LocalizationUncertainty(model=2).compute(locdata_simple)
        assert all(key_ in unc.results.columns for key_ in ["uncertainty_x"])
        assert unc.results.iloc[-1, 0] == pytest.approx(100, rel=0.01)

    def test_kwargs(self, locdata_simple):
        with pytest.warns(RuntimeWarning):
            unc = LocalizationUncertainty(model=2, psf_sigma_x=130).compute(
                locdata_simple
            )
        assert all(key_ in unc.results.columns for key_ in ["uncertainty_x"])
        assert unc.results.iloc[-1, 0] == pytest.approx(130, rel=0.01)

        with pytest.warns(RuntimeWarning):
            unc = LocalizationUncertainty(
                model=localization_precision_model_3, pixel_size=130
            ).compute(locdata_simple)
        assert all(key_ in unc.results.columns for key_ in ["uncertainty_x"])
        assert unc.results.iloc[-1, 0] == pytest.approx(192.8, rel=0.01)



"""
.. currentmodule:: locan

Locan consists of the following modules:

.. autosummary::
   :toctree: generated/

   analysis
   constants
   configuration
   data
   datasets
   dependencies
   gui
   locan_io
   process
   rois
   scripts
   simulation
   tests
   utils
   visualize
"""

from __future__ import annotations

import lazy_loader as lazy
import logging

logging.getLogger(__name__).addHandler(logging.NullHandler())

try:
    from locan._version import version as __version__
except ImportError:
    __version__ = "not-installed"


__getattr__, __lazy_dir__, __lazy_all__ = lazy.attach_stub(
    package_name=__name__, filename=__file__
)

__all__ = __lazy_all__ + ["__version__", "locdata_id"]


# Identifier for LocData objects that is reset for each locan session and incremented with each LocData instantiation.
locdata_id: int = 0


def __dir__():
    return __lazy_dir__() + ["__version__", "locdata_id"]


# initialize configuration settings
import locan.configuration



"""
Analyze cross dependencies between localization properties.

Analyze cross dependencies as indicated by the correlation coefficients between any two localization properties.
"""

from __future__ import annotations

import logging
import sys
from typing import TYPE_CHECKING, Any

import pandas as pd

from locan import get_colormap

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

if TYPE_CHECKING:
    import matplotlib as mpl

    from locan.data.locdata import LocData

import matplotlib.pyplot as plt
import numpy as np

from locan.analysis import metadata_analysis_pb2
from locan.analysis.analysis_base import _Analysis
from locan.configuration import COLORMAP_DEFAULTS

__all__: list[str] = ["LocalizationPropertyCorrelations"]

logger = logging.getLogger(__name__)


# The algorithms


def _localization_property_correlations(
    locdata: LocData, loc_properties: list[str] | None = None
) -> pd.DataFrame:
    if loc_properties is None:
        results = locdata.data.corr()
    else:
        results = locdata.data[loc_properties].corr()
    return results


# The specific analysis classes


class LocalizationPropertyCorrelations(_Analysis):
    """
    Compute and analyze correlation coefficients between any two localization
     properties.

    Parameters
    ----------
    meta : locan.analysis.metadata_analysis_pb2.AMetadata
        Metadata about the current analysis routine.
    loc_properties : list[str] | None
        Localization properties to be analyzed. If None all are used.

    Attributes
    ----------
    count : int
        A counter for counting instantiations (class attribute).
    parameter : dict[str, Any]
        A dictionary with all settings for the current computation.
    meta : locan.analysis.metadata_analysis_pb2.AMetadata | None
        Metadata about the current analysis routine.
    results : pandas.DataFrame | None
        The correlation coefficients..
    """

    def __init__(
        self,
        meta: metadata_analysis_pb2.AMetadata | None = None,
        loc_properties: list[str] | None = None,
    ) -> None:
        parameters = self._get_parameters(locals())
        super().__init__(**parameters)
        self.results: pd.DataFrame | None = None

    def compute(self, locdata: LocData) -> Self:
        """
        Run the computation.

        Parameters
        ----------
        locdata
            Localization data.

        Returns
        -------
        Self
        """
        if not len(locdata):
            logger.warning("Locdata is empty.")
            return self

        self.results = _localization_property_correlations(
            locdata=locdata, **self.parameter
        )
        return self

    def report(self) -> None:
        if self.results is None:
            logger.warning("No results available")
            return

        print("Fit results for:\n")
        print(self.results.model_result.fit_report(min_correl=0.25))
        # print(self.results.fit_results.best_values)

    def plot(
        self,
        ax: mpl.axes.Axes | None = None,
        cbar: bool = True,
        colorbar_kws: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> mpl.axes.Axes:
        """
        Provide heatmap of all correlation values as
        :class:`matplotlib.axes.Axes` object.

        Parameters
        ----------
        ax
            The axes on which to show the image
        cbar
            If true draw a colorbar.
        colorbar_kws
            Keyword arguments for :func:`matplotlib.pyplot.colorbar`.
        kwargs
            Other parameters passed to :func:`matplotlib.pyplot.imshow`.

        Returns
        -------
        matplotlib.axes.Axes
            Axes object with the plot.
        """
        if ax is None:
            ax = plt.gca()

        if self.results is None:
            return ax

        im = ax.imshow(
            self.results,
            **dict(
                dict(
                    vmin=-1,
                    vmax=1,
                    cmap=get_colormap(
                        colormap=COLORMAP_DEFAULTS["DIVERGING"]
                    ).matplotlib,
                ),
                **kwargs,
            ),
        )
        columns = self.results.columns

        # We want to show all ticks...
        ax.set_xticks(np.arange(len(columns)))
        ax.set_yticks(np.arange(len(columns)))

        # ensure correct scaling
        ax.set_xticks(np.arange(len(columns) + 1) - 0.5, minor=True)
        ax.set_yticks(np.arange(len(columns) + 1) - 0.5, minor=True)
        ax.tick_params(which="minor", bottom=False, left=False)

        # ... and label them with the respective list entries
        ax.set_xticklabels(columns)
        ax.set_yticklabels(columns)

        # Rotate the tick labels and set their alignment.
        plt.setp(ax.get_xticklabels(), rotation=45, ha="right", rotation_mode="anchor")

        # Loop over data dimensions and create text annotations.
        for i in range(len(columns)):
            for j in range(len(columns)):
                ax.text(
                    j,
                    i,
                    round(self.results.values[i, j], 2),
                    ha="center",
                    va="center",
                    color="w",
                )

        ax.set_title("Localization Property Correlations")

        # Create colorbar
        if cbar:
            if colorbar_kws is None:
                cbar_ = ax.figure.colorbar(im)  # type:ignore[union-attr]
            else:
                cbar_ = ax.figure.colorbar(  # type:ignore[union-attr]  # noqa: F841
                    im, **colorbar_kws
                )
            # cbar_.ax.set_ylabel('correlation', rotation=-90, va="bottom")

        ax.figure.tight_layout()  # type:ignore[union-attr]
        return ax



import pytest

from locan import (
    render_3d_napari,
    render_3d_napari_image,
    render_3d_rgb_napari,
    transform_affine,
)
from locan.dependencies import HAS_DEPENDENCY

napari = pytest.importorskip("napari")

HAS_NAPARI_AND_PYTESTQT = HAS_DEPENDENCY["napari"] and HAS_DEPENDENCY["pytestqt"]
# pytestqt is not a requested or extra dependency.
# If napari and pytest-qt is installed, all tests run.
# Tests in docker or GitHub actions on linux require xvfb
# for tests with pytest-qt to run.


@pytest.mark.visual
# this is to check overlay of rendered image and single localization points
def test_render_3d_napari_coordinates(locdata_blobs_3d):
    viewer = render_3d_napari(locdata_blobs_3d, bin_size=10, cmap="viridis", gamma=0.1)
    viewer.add_points(locdata_blobs_3d.coordinates)
    napari.run()


def test_render_3d_napari_image(locdata_blobs_3d):
    with pytest.raises(ValueError):
        render_3d_napari_image(
            locdata_blobs_3d, bin_edges=((0, 10, 100), (0, 10, 100), (0, 10, 100))
        )

    data, image_kwargs, layer_type = render_3d_napari_image(
        locdata_blobs_3d, bin_size=100, cmap="viridis", gamma=0.1
    )
    assert layer_type == "image"
    assert all(
        key in ["name", "colormap", "scale", "translate", "metadata", "gamma"]
        for key in image_kwargs
    )
    assert data.shape == (7, 6, 8)


@pytest.mark.skipif(
    not HAS_NAPARI_AND_PYTESTQT, reason="Test requires napari and pytest-qt."
)
def test_render_3d_napari(make_napari_viewer, locdata_blobs_3d):
    viewer = make_napari_viewer()
    viewer_ = render_3d_napari(
        locdata_blobs_3d,
        viewer=viewer,
        bin_size=100,
        cmap="cyan",
        gamma=0.1,
        scale=(2, 2),
        blending="additive",
    )
    assert viewer_ is viewer


@pytest.mark.skipif(
    not HAS_NAPARI_AND_PYTESTQT, reason="Test requires napari and pytest-qt."
)
def test_render_3d_napari_empty(make_napari_viewer, locdata_empty):
    viewer = make_napari_viewer()
    render_3d_napari(
        locdata_empty, viewer=viewer, bin_size=100, cmap="viridis", gamma=0.1
    )
    assert viewer.layers == []


@pytest.mark.gui
def test_render_3d_napari_single_gui(locdata_single_localization_3d, caplog):
    render_3d_napari(
        locdata_single_localization_3d, bin_size=100, cmap="viridis", gamma=0.1
    )
    try:
        record_tuples_ = caplog.record_tuples[1]
    except IndexError:
        record_tuples_ = caplog.record_tuples[0]
    assert record_tuples_ == (
        "locan.visualize.render_napari.render3d",
        30,
        "Locdata carries a single localization.",
    )
    napari.run()


@pytest.mark.skipif(
    not HAS_NAPARI_AND_PYTESTQT, reason="Test requires napari and pytest-qt."
)
def test_render_3d_napari_single(
    make_napari_viewer, locdata_single_localization_3d, caplog
):
    viewer = make_napari_viewer()
    render_3d_napari(
        locdata_single_localization_3d,
        viewer=viewer,
        bin_size=100,
        cmap="viridis",
        gamma=0.1,
    )
    assert caplog.record_tuples[-1] == (
        "locan.visualize.render_napari.render3d",
        30,
        "Locdata carries a single localization.",
    )


@pytest.mark.gui
def test_render_3d_rgb_napari_gui(locdata_blobs_3d):
    locdata_0 = locdata_blobs_3d
    locdata_1 = transform_affine(locdata_blobs_3d, offset=(20, 0, 0))
    render_3d_rgb_napari([locdata_0, locdata_1], bin_size=20)

    napari.run()


@pytest.mark.skipif(
    not HAS_NAPARI_AND_PYTESTQT, reason="Test requires napari and pytest-qt."
)
def test_render_3d_rgb_napari(make_napari_viewer, locdata_blobs_3d):
    viewer = make_napari_viewer()
    locdata_0 = locdata_blobs_3d
    locdata_1 = transform_affine(locdata_blobs_3d, offset=(20, 0, 0))
    render_3d_rgb_napari([locdata_0, locdata_1], viewer=viewer, bin_size=20)



import pytest
from matplotlib import colors as mcolors

from locan.configuration import COLORMAP_DEFAULTS
from locan.dependencies import HAS_DEPENDENCY
from locan.visualize import (
    Colormap,
    Colormaps,
    colormap_registry,
    get_colormap,
)

if HAS_DEPENDENCY["napari"]:
    import napari


def test_COLORMAP_DEFAULTS_and_Colormaps():
    assert "CONTINUOUS" in COLORMAP_DEFAULTS
    assert all(hasattr(Colormaps, key) for key in COLORMAP_DEFAULTS.keys())
    assert all(
        name in COLORMAP_DEFAULTS for name, value in Colormaps.__members__.items()
    )
    assert COLORMAP_DEFAULTS[Colormaps.CONTINUOUS.name] == "cet_fire"
    assert COLORMAP_DEFAULTS[Colormaps.CONTINUOUS_REVERSE.name] == "cet_fire_r"
    assert COLORMAP_DEFAULTS[Colormaps.CONTINUOUS_GRAY.name] == "cet_gray"
    assert COLORMAP_DEFAULTS[Colormaps.CONTINUOUS_GRAY_REVERSE.name] == "cet_gray_r"
    assert COLORMAP_DEFAULTS[Colormaps.DIVERGING.name] == "cet_coolwarm"
    assert COLORMAP_DEFAULTS[Colormaps.CATEGORICAL.name] == "cet_glasbey_dark"
    assert COLORMAP_DEFAULTS[Colormaps.TURBO.name] == "turbo"


def test_colormap_registry():
    assert all(value in colormap_registry for value in COLORMAP_DEFAULTS.values())


class TestColormap:
    def test_init(self):
        with pytest.raises(TypeError):
            Colormap()

        colormap = Colormap(colormap=mcolors.Colormap("viridis"))
        assert isinstance(colormap.matplotlib, mcolors.Colormap)
        assert colormap.name == "viridis"

    @pytest.mark.skipif(not HAS_DEPENDENCY["napari"], reason="Test requires napari.")
    def test_init_with_napari(self):
        white_green_cmap = {
            "colors": [[1, 1, 1, 1], [0, 1, 0, 1]],
            "name": "white_to_green",
            "interpolation": "linear",
        }
        colormap = napari.utils.colormaps.Colormap(**white_green_cmap)
        colormap = Colormap(colormap=colormap)
        assert isinstance(colormap.napari, napari.utils.Colormap)
        assert colormap.name == "white_to_green"

    def test_from_matplotlib(self):
        colormap = Colormap.from_matplotlib(colormap="viridis")
        assert isinstance(colormap.matplotlib, mcolors.Colormap)
        assert colormap.name == "viridis"

        colormap = Colormap.from_matplotlib(colormap=mcolors.Colormap("viridis"))
        assert isinstance(colormap.matplotlib, mcolors.Colormap)
        assert colormap.name == "viridis"

    @pytest.mark.skipif(not HAS_DEPENDENCY["napari"], reason="Test requires napari.")
    def test_from_matplotlib_to_napari(self):
        colormap = Colormap.from_matplotlib(colormap="viridis")
        assert isinstance(colormap.matplotlib, mcolors.Colormap)
        assert colormap.name == "viridis"
        assert isinstance(colormap.napari, napari.utils.Colormap)
        assert colormap.napari.name == "viridis"

        # from locan default colormaps that are not in napari
        colormap = colormap_registry["cet_fire"]
        assert isinstance(colormap.napari, napari.utils.Colormap)
        assert colormap.name == "cet_fire"

    @pytest.mark.skipif(not HAS_DEPENDENCY["napari"], reason="Test requires napari.")
    def test_from_napari(self):
        white_green_cmap = {
            "colors": [[1, 1, 1, 1], [0, 1, 0, 1]],
            "name": "white_to_green",
            "interpolation": "linear",
        }
        colormap_ = napari.utils.colormaps.Colormap(**white_green_cmap)

        colormap = Colormap.from_napari(colormap=colormap_)
        assert isinstance(colormap.napari, napari.utils.Colormap)
        assert colormap.name == "white_to_green"
        assert isinstance(colormap.matplotlib, mcolors.Colormap)
        assert colormap.name == "white_to_green"

        colormap = Colormap.from_napari(colormap=white_green_cmap)
        assert isinstance(colormap.napari, napari.utils.Colormap)
        assert colormap.name == "white_to_green"

        colormap = Colormap.from_napari(colormap="viridis")
        assert isinstance(colormap.napari, napari.utils.Colormap)
        assert colormap.name == "viridis"

        with pytest.raises(LookupError):
            Colormap.from_napari(colormap="not in napari")

    @pytest.mark.skipif(
        not HAS_DEPENDENCY["colorcet"], reason="Test requires colorcet."
    )
    def test_from_colorcet(self):
        colormap = Colormap.from_colorcet(colormap="cet_fire")
        assert isinstance(colormap.matplotlib, mcolors.Colormap)
        assert colormap.name == "cet_fire"

        colormap = Colormap.from_colorcet(colormap="fire")
        assert isinstance(colormap.matplotlib, mcolors.Colormap)
        assert colormap.name == "fire"

    def test_from_registry(self):
        colormap = Colormap.from_registry(colormap="cet_fire")
        assert isinstance(colormap.matplotlib, mcolors.Colormap)
        assert colormap.name == "cet_fire"

        colormap = Colormap.from_registry(colormap="viridis")
        assert isinstance(colormap.matplotlib, mcolors.Colormap)
        assert colormap.name == "viridis"

        colormap = Colormap.from_registry(colormap=Colormaps.TURBO.value)
        assert isinstance(colormap.matplotlib, mcolors.Colormap)
        assert colormap.name == "turbo"

        colormap = Colormap.from_registry(colormap=COLORMAP_DEFAULTS["TURBO"])
        assert isinstance(colormap.matplotlib, mcolors.Colormap)
        assert colormap.name == "turbo"

        with pytest.raises(LookupError):
            Colormap.from_registry(colormap="not in registry")

    def test_call(self):
        colormap = Colormap.from_matplotlib(colormap="viridis")
        assert len(colormap(0.5)) == 4
        assert len(colormap([0, 0.5, 1])) == 3

    @pytest.mark.skipif(not HAS_DEPENDENCY["napari"], reason="Test requires napari.")
    def test_call_napari(self):
        colormap = Colormap.from_napari(colormap="viridis")
        assert colormap(0.5).shape == (1, 4)
        # assert colormap(0.5).tolist()[0] == [0.12814850801305894, 0.5651070037572249, 0.5508925046267017, 1.0]
        assert colormap([0, 0.5, 1]).shape == (3, 4)


def test_get_colormap():
    colormap = Colormap.from_matplotlib(colormap="viridis")
    colormap = get_colormap(colormap)
    assert isinstance(colormap.matplotlib, mcolors.Colormap)
    assert colormap.name == "viridis"

    colormap = get_colormap("viridis")
    assert isinstance(colormap.matplotlib, mcolors.Colormap)
    assert colormap.name == "viridis"

    colormap = get_colormap(Colormaps.TURBO)
    assert isinstance(colormap.matplotlib, mcolors.Colormap)
    assert colormap.name == "turbo"

    colormap = get_colormap(Colormaps.TURBO.value)
    assert isinstance(colormap.matplotlib, mcolors.Colormap)
    assert colormap.name == "turbo"

    with pytest.raises(TypeError):
        get_colormap("does not exist")
    with pytest.raises(TypeError):
        get_colormap([])


def test_colormap_definitions():
    colormap = get_colormap("cet_fire")
    assert isinstance(colormap.matplotlib, mcolors.Colormap)
    assert colormap.name == "cet_fire"

    colormap = get_colormap("cet_fire_r")
    assert isinstance(colormap.matplotlib, mcolors.Colormap)
    assert colormap.name == "cet_fire_r"

    colormap = get_colormap("cet_gray")
    assert isinstance(colormap.matplotlib, mcolors.Colormap)
    assert colormap.name == "cet_gray"

    colormap = get_colormap("cet_gray_r")
    assert isinstance(colormap.matplotlib, mcolors.Colormap)
    assert colormap.name == "cet_gray_r"

    colormap = get_colormap("cet_glasbey_dark")
    assert isinstance(colormap.matplotlib, mcolors.Colormap)
    assert colormap.name == "cet_glasbey_dark"



# -*- coding: utf-8 -*-

# Configuration file for the Sphinx documentation builder.
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import os
import sys
from importlib.metadata import version as get_version
from typing import Any

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here.
sys.path.insert(0, os.path.abspath(".."))


# -- Project information -----------------------------------------------------

# General information about the project.
project = "Locan"
author = "Locan Developers"
copyright = "2018-2025, Locan Developers"

# Version number generated by setuptools_scm
release: str = get_version("locan")
# take major/minor
# version = '.'.join(release.split('.')[:2])
version: str = release


# -- General configuration ---------------------------------------------------

# Minimal Sphinx version needed for documentation
# needs_sphinx = '1.0'

# Sphinx extension modules
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",
    "sphinx.ext.inheritance_diagram",
    "sphinx.ext.intersphinx",
    "sphinx.ext.mathjax",
    "sphinx.ext.viewcode",
    "IPython.sphinxext.ipython_console_highlighting",
    "IPython.sphinxext.ipython_directive",
    "myst_nb",
    "sphinx_autodoc_typehints",
    "sphinx_copybutton",
    # 'sphinx_rtd_theme',
]

# autosummary settings
# Make _autosummary files and include them
autosummary_generate = True

# Napoleon settings
napoleon_google_docstring = True
napoleon_numpy_docstring = True
napoleon_include_init_with_doc = True
napoleon_include_private_with_doc = False
napoleon_include_special_with_doc = False
napoleon_use_admonition_for_examples = False
napoleon_use_admonition_for_notes = False
napoleon_use_admonition_for_references = False
napoleon_use_ivar = True
napoleon_use_param = True
napoleon_use_rtype = True
napoleon_preprocess_types = False
napoleon_type_aliases = None
napoleon_attr_annotations = True

# myST-NB settings
nb_execution_mode = "force"  # 'off' 'cache' 'force'

# copybutton settings
copybutton_prompt_text = ">>> "

# The language for content autogenerated by Sphinx.
language = "en"

# The suffixes of source filenames.
source_suffix = {
    ".rst": "restructuredtext",
    ".md": "myst-nb",
    ".ipynb": "myst-nb",
    ".myst": "myst-nb",
}

# The master/root toctree document.
master_doc = root_doc = "index"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = [
    "_build",
    "jupyter_execute",
    ".jupyter_cache",
    "**.ipynb_checkpoints",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.
# html_theme = "sphinx_rtd_theme"
html_theme = "furo"

# Theme options
# for furo
html_theme_options: dict[str, Any] = {
    # 'sidebar_hide_name': True,
}

html_title = f"{project}\n{release}"
html_short_title = f"{project}"

html_logo = "_static/logo.png"
html_favicon = "_static/favicon.ico"

html_static_path = ["_static"]


# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = f"{project}_docs"


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    "papersize": "a4paper",
    # The font size ('10pt', '11pt' or '12pt').
    "pointsize": "11pt",
    # Additional stuff for the LaTeX preamble.
    "preamble": r"\setcounter{tocdepth}{2}",
    # Latex figure (float) alignment
    # 'figure_align': 'htbp',
    "classoptions": ",openany,oneside",
}

latex_logo = "_static/logo.png"

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, "Locan.tex", "Locan Documentation", author, "manual"),
]


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, "locan", "Locan Documentation", [author], 1)]


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        "Locan",
        "Locan Documentation",
        author,
        "Locan",
        "One line description of project.",
        "Miscellaneous",
    ),
]



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: locan/analysis/metadata_analysis.proto
# Protobuf Python Version: 5.28.2
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    28,
    2,
    "",
    "locan/analysis/metadata_analysis.proto",
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n&locan/analysis/metadata_analysis.proto\x12\x0elocan.analysis\x1a\x1fgoogle/protobuf/timestamp.proto"3\n\x10\x41nalysis_routine\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x11\n\tparameter\x18\x02 \x01(\t"\xd7\x02\n\tAMetadata\x12\x12\n\nidentifier\x18\x01 \x01(\t\x12\x0f\n\x07\x63omment\x18\x02 \x01(\t\x12\x30\n\x06method\x18\x05 \x01(\x0b\x32 .locan.analysis.Analysis_routine\x12/\n\x03map\x18\x07 \x03(\x0b\x32".locan.analysis.AMetadata.MapEntry\x12\x31\n\rcreation_time\x18\x08 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x35\n\x11modification_time\x18\t \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x1a*\n\x08MapEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01J\x04\x08\x03\x10\x04J\x04\x08\x04\x10\x05R\rcreation_dateR\x11modification_date'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(
    DESCRIPTOR, "locan.analysis.metadata_analysis_pb2", _globals
)
if not _descriptor._USE_C_DESCRIPTORS:
    DESCRIPTOR._loaded_options = None
    _globals["_AMETADATA_MAPENTRY"]._loaded_options = None
    _globals["_AMETADATA_MAPENTRY"]._serialized_options = b"8\001"
    _globals["_ANALYSIS_ROUTINE"]._serialized_start = 91
    _globals["_ANALYSIS_ROUTINE"]._serialized_end = 142
    _globals["_AMETADATA"]._serialized_start = 145
    _globals["_AMETADATA"]._serialized_end = 488
    _globals["_AMETADATA_MAPENTRY"]._serialized_start = 400
    _globals["_AMETADATA_MAPENTRY"]._serialized_end = 442
# @@protoc_insertion_point(module_scope)



"""
Compute localization precision from successive nearby localizations.

Localization precision is estimated from spatial variations of all
localizations that appear in successive frames
within a specified search radius [1]_.

Localization pair distance distributions are fitted according to the
probability density functions in [2]_.

The estimated sigmas describe the standard deviation for pair distances.
Localization precision is often defined as the standard deviation for
localization distances from the center position.
With that definition, the localization precision is equal to sigma / sqrt(2).

References
----------
.. [1] Endesfelder, Ulrike, et al.,
   A simple method to estimate the average localization precision of a
   single-molecule localization microscopy experiment.
   Histochemistry and Cell Biology 141.6 (2014): 629-638.

.. [2] L. Stirling Churchman, Henrik Flyvbjerg, James A. Spudich,
   A Non-Gaussian Distribution Quantifies Distances Measured with Fluorescence
   Localization Techniques.
   Biophysical Journal 90 (2), 2006, 668-671,
   doi.org/10.1529/biophysj.105.065599.
"""

from __future__ import annotations

import logging
import sys
from collections.abc import Sequence
from typing import TYPE_CHECKING, Any

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

if TYPE_CHECKING:
    import matplotlib as mpl

    from locan.data.locdata import LocData

import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
import pandas as pd
from scipy import stats
from sklearn.neighbors import NearestNeighbors
from tqdm import tqdm

from locan.analysis import metadata_analysis_pb2
from locan.analysis.analysis_base import _Analysis
from locan.configuration import N_JOBS, TQDM_DISABLE, TQDM_LEAVE

__all__: list[str] = ["LocalizationPrecision"]

logger = logging.getLogger(__name__)


# The algorithms


def _localization_precision(locdata: LocData, radius: int | float = 50) -> pd.DataFrame:
    # group localizations
    grouped = locdata.data.groupby("frame")

    # find nearest neighbors
    min = locdata.data["frame"].unique().min()
    max = locdata.data["frame"].unique().max()

    results = pd.DataFrame()

    for i in tqdm(
        range(min, max - 1),
        desc="Processed frames:",
        leave=TQDM_LEAVE,
        disable=TQDM_DISABLE,
    ):
        try:
            points = grouped.get_group(i)[locdata.coordinate_keys]
            other_points = grouped.get_group(i + 1)[locdata.coordinate_keys]

            nn = NearestNeighbors(radius=radius, metric="euclidean", n_jobs=N_JOBS).fit(
                other_points
            )
            distances, indices = nn.radius_neighbors(points)

            if len(distances):
                for n, (dists, inds) in enumerate(zip(distances, indices)):
                    if len(dists):
                        min_distance = np.amin(dists)
                        min_position = np.argmin(dists)
                        min_index = inds[min_position]
                        difference = points.iloc[n] - other_points.iloc[min_index]

                        df = difference.to_frame().T
                        df = df.rename(
                            columns={
                                "position_x": "position_delta_x",
                                "position_y": "position_delta_y",
                                "position_z": "position_delta_z",
                            }
                        )
                        df = df.assign(position_distance=min_distance)
                        df = df.assign(frame=i)
                        results = pd.concat([results, df])
        except KeyError:
            pass

    results.reset_index(inplace=True, drop=True)
    return results


# The specific analysis classes


class LocalizationPrecision(_Analysis):
    """
    Compute the localization precision from consecutive nearby localizations.

    Parameters
    ----------
    meta : locan.analysis.metadata_analysis_pb2.AMetadata | None
        Metadata about the current analysis routine.
    radius : int | float
        Search radius for nearest-neighbor searches.

    Attributes
    ----------
    count : int
        A counter for counting instantiations (class attribute).
    parameter : dict
        A dictionary with all settings for the current computation.
    meta : locan.analysis.metadata_analysis_pb2.AMetadata
        Metadata about the current analysis routine.
    results : pandas.DataFrame
        Computed results.
    distribution_statistics : Distribution_fits | None
        Distribution parameters derived from MLE fitting of results.
    """

    def __init__(
        self,
        meta: metadata_analysis_pb2.AMetadata | None = None,
        radius: int | float = 50,
    ) -> None:
        parameters = self._get_parameters(locals())
        super().__init__(**parameters)
        self.results = None
        self.distribution_statistics: _DistributionFits | None = None

    def compute(self, locdata: LocData) -> Self:
        """
        Run the computation.

        Parameters
        ----------
        locdata
            Localization data.

        Returns
        -------
        Self
        """
        if not len(locdata):
            logger.warning("Locdata is empty.")
            return self

        self.results = _localization_precision(locdata=locdata, **self.parameter)
        if self.results.empty:
            logger.warning("No successive localizations were found.")

        return self

    def fit_distributions(self, loc_property: str | None = None, **kwargs: Any) -> None:
        """
        Fit probability density functions to the distributions of
        `loc_property` values in the results using MLE (scipy.stats).

        Parameters
        ----------
        loc_property : str
            The LocData property for which to fit an appropriate distribution;
            if None all plots are shown.
        kwargs
            Other parameters passed to the `distribution.fit()` method.
        """
        if self.results is None:
            logger.warning("No results available to be fitted.")
            return

        self.distribution_statistics = _DistributionFits(self)

        if loc_property is None:
            for prop in [
                "position_delta_x",
                "position_delta_y",
                "position_delta_z",
                "position_distance",
            ]:
                if prop in self.results.columns:
                    self.distribution_statistics.fit(loc_property=prop, **kwargs)
        else:
            self.distribution_statistics.fit(loc_property=loc_property, **kwargs)

    def plot(
        self,
        ax: mpl.axes.Axes | None = None,
        loc_property: str | list[str] | None = None,
        window: int = 1,
        **kwargs: Any,
    ) -> mpl.axes.Axes:
        """
        Provide plot as :class:`matplotlib.axes.Axes` object showing the
        running average of results over window size.

        Parameters
        ----------
        ax
            The axes on which to show the image
        loc_property
            The property for which to plot localization precision;
            if None all plots are shown.
        window
            Window for running average that is applied before plotting.
        kwargs
            Other parameters passed to :func:`matplotlib.pyplot.plot`.

        Returns
        -------
        matplotlib.axes.Axes
            Axes object with the plot.
        """
        if ax is None:
            ax = plt.gca()

        if self.results is None:
            return ax

        # prepare plot
        self.results.rolling(window=window, center=True).mean().plot(
            ax=ax, x="frame", y=loc_property, **dict(dict(legend=False), **kwargs)
        )
        ax.set(
            title=f"Localization Precision\n (window={window})",
            xlabel="frame",
            ylabel=loc_property,
        )

        return ax

    def hist(
        self,
        ax: mpl.axes.Axes | None = None,
        loc_property: str = "position_distance",
        bins: int | Sequence[int | float] | str = "auto",
        fit: bool = True,
        **kwargs: Any,
    ) -> mpl.axes.Axes:
        """
        Provide histogram as :class:`matplotlib.axes.Axes` object showing the
        distributions of results.

        Parameters
        ----------
        ax
            The axes on which to show the image
        loc_property
            The property for which to plot localization precision.
        bins
            Bin specifications (passed to :func:`matplotlib.hist`).
        fit
            Flag indicating if distributions fit are shown.
        kwargs
            Other parameters passed to :func:`matplotlib.pyplot.his`.

        Returns
        -------
        matplotlib.axes.Axes
            Axes object with the plot.
        """
        if ax is None:
            ax = plt.gca()

        if self.results is None:
            return ax

        # prepare plot
        ax.hist(
            self.results[loc_property].values,
            bins=bins,
            **dict(dict(density=True, log=False), **kwargs),
        )
        ax.set(title="Localization Precision", xlabel=loc_property, ylabel="PDF")

        if fit:
            if self.distribution_statistics is None:
                self.fit_distributions()

            assert (  # type narrowing # noqa: S101
                self.distribution_statistics is not None
            )
            self.distribution_statistics.plot(ax=ax, loc_property=loc_property)

        return ax


# Auxiliary functions and classes


class PairwiseDistance1d(stats.rv_continuous):
    """
    A random variable describing the distribution
    of pair distances for two normal distributed point clouds in 1D.

    The continuous distribution class inherits from scipy.stats.rv_continuous
    a set of methods and is defined by overriding the _pdf method.

    For theoretical background see [1]_.

    Parameters
    ----------
    x : npt.ArrayLike
        distance
    mu : npt.ArrayLike
        Distance between the point cloud center positions.
    sigma_1 : npt.ArrayLike
        Standard deviation for one point cloud.
    sigma_2 : npt.ArrayLike
        Standard deviation for the other point cloud.

    References
    ----------
    .. [1] L. Stirling Churchman, Henrik Flyvbjerg, James A. Spudich,
       A Non-Gaussian Distribution Quantifies Distances Measured with
       Fluorescence Localization Techniques.
       Biophysical Journal 90 (2), 2006, 668-671,
       doi.org/10.1529/biophysj.105.065599.
    """

    def _pdf(
        self,
        x: npt.ArrayLike,
        mu: npt.ArrayLike,
        sigma_1: npt.ArrayLike,
        sigma_2: npt.ArrayLike,
    ) -> npt.NDArray[np.float64]:
        x = np.asarray(x)
        mu = np.asarray(mu)
        sigma_1 = np.asarray(sigma_1)
        sigma_2 = np.asarray(sigma_2)

        sigma = np.sqrt(sigma_1**2 + sigma_2**2)
        return_value: npt.NDArray[np.float64] = (
            np.sqrt(2 / np.pi)
            / sigma
            * np.exp(-(mu**2 + x**2) / (2 * sigma**2))
            * np.cosh(x * mu / (sigma**2))
        )
        return return_value


class PairwiseDistance2d(stats.rv_continuous):
    """
    A random variable describing the distribution
    of pair distances for two normal distributed point clouds in 2D.

    The continuous distribution class inherits from scipy.stats.rv_continuous
    a set of methods and is defined by overriding the _pdf method.

    For theoretical background see [1]_.

    Parameters
    ----------
    x : npt.ArrayLike
        distance
    mu : npt.ArrayLike
        Distance between the point cloud center positions.
    sigma_1 : npt.ArrayLike
        Standard deviation for one point cloud.
    sigma_2 : npt.ArrayLike
        Standard deviation for the other point cloud.

    References
    ----------
    .. [1] L. Stirling Churchman, Henrik Flyvbjerg, James A. Spudich,
       A Non-Gaussian Distribution Quantifies Distances Measured with
       Fluorescence Localization Techniques.
       Biophysical Journal 90 (2), 2006, 668-671,
       doi.org/10.1529/biophysj.105.065599.
    """

    def _pdf(
        self,
        x: npt.ArrayLike,
        mu: npt.ArrayLike,
        sigma_1: npt.ArrayLike,
        sigma_2: npt.ArrayLike,
    ) -> npt.NDArray[np.float64]:
        x = np.asarray(x)
        mu = np.asarray(mu)
        sigma_1 = np.asarray(sigma_1)
        sigma_2 = np.asarray(sigma_2)

        sigma = np.sqrt(sigma_1**2 + sigma_2**2)
        return_value: npt.NDArray[np.float64] = (
            x
            / (sigma**2)
            * np.exp(-(mu**2 + x**2) / (2 * sigma**2))
            * np.i0(x * mu / (sigma**2))
        )
        return return_value


class PairwiseDistance2dIdenticalSigma(stats.rv_continuous):
    """
    A random variable describing the distribution
    of pair distances for two normal distributed point clouds in 2D.

    The continuous distribution class inherits from scipy.stats.rv_continuous
    a set of methods and is defined by overriding the _pdf method.

    For theoretical background see [1]_.

    Parameters
    ----------
    x : npt.ArrayLike
        distance
    mu : npt.ArrayLike
        Distance between the point cloud center positions.
    sigma : npt.ArrayLike
        Standard deviation for both point clouds.

    References
    ----------
    .. [1] L. Stirling Churchman, Henrik Flyvbjerg, James A. Spudich,
       A Non-Gaussian Distribution Quantifies Distances Measured with
       Fluorescence Localization Techniques.
       Biophysical Journal 90 (2), 2006, 668-671,
       doi.org/10.1529/biophysj.105.065599.
    """

    def _pdf(
        self, x: npt.ArrayLike, mu: npt.ArrayLike, sigma: npt.ArrayLike
    ) -> npt.NDArray[np.float64]:
        x = np.asarray(x)
        mu = np.asarray(mu)
        sigma = np.asarray(sigma)

        return_value: npt.NDArray[np.float64] = (
            x
            / (sigma**2)
            * np.exp(-(mu**2 + x**2) / (2 * sigma**2))
            * np.i0(x * mu / (sigma**2))
        )
        return return_value


class PairwiseDistance3d(stats.rv_continuous):
    """
    A random variable describing the distribution
    of pair distances for two normal distributed point clouds in 3D.

    The continuous distribution class inherits from scipy.stats.rv_continuous
    a set of methods and is defined by overriding the _pdf method.

    For theoretical background see [1]_.

    Parameters
    ----------
    x : npt.ArrayLike
        distance
    mu : npt.ArrayLike
        Distance between the point cloud center positions.
    sigma_1 : npt.ArrayLike
        Standard deviation for one point cloud.
    sigma_2 : npt.ArrayLike
        Standard deviation for the other point cloud.

    References
    ----------
    .. [1] L. Stirling Churchman, Henrik Flyvbjerg, James A. Spudich,
       A Non-Gaussian Distribution Quantifies Distances Measured with
       Fluorescence Localization Techniques.
       Biophysical Journal 90 (2), 2006, 668-671,
       doi.org/10.1529/biophysj.105.065599.
    """

    def _pdf(
        self,
        x: npt.ArrayLike,
        mu: npt.ArrayLike,
        sigma_1: npt.ArrayLike,
        sigma_2: npt.ArrayLike,
    ) -> npt.NDArray[np.float64]:
        x = np.asarray(x)
        mu = np.asarray(mu)
        sigma_1 = np.asarray(sigma_1)
        sigma_2 = np.asarray(sigma_2)

        sigma = np.sqrt(sigma_1**2 + sigma_2**2)
        if all(mu == 0):
            return_value: npt.NDArray[np.float64] = (
                np.sqrt(2 / np.pi)
                * x
                / sigma
                * np.exp(-(mu**2 + x**2) / (2 * sigma**2))
                * x
                / (sigma**2)
            )
        else:
            return_value = (
                np.sqrt(2 / np.pi)
                * x
                / sigma
                / mu
                * np.exp(-(mu**2 + x**2) / (2 * sigma**2))
                * np.sinh(x * mu / (sigma**2))
            )
        return return_value


class PairwiseDistance1dIdenticalSigmaZeroMu(stats.rv_continuous):
    """
    A random variable describing the distribution
    of pair distances for two normal distributed point clouds in 1D.

    The continuous distribution class inherits from scipy.stats.rv_continuous
    a set of methods and is defined by overriding the _pdf method.

    For theoretical background see [1]_.

    Parameters
    ----------
    x : npt.ArrayLike
        distance
    sigma : npt.ArrayLike
        Standard deviation for point clouds

    References
    ----------
    .. [1] L. Stirling Churchman, Henrik Flyvbjerg, James A. Spudich,
       A Non-Gaussian Distribution Quantifies Distances Measured with
       Fluorescence Localization Techniques.
       Biophysical Journal 90 (2), 2006, 668-671,
       doi.org/10.1529/biophysj.105.065599.
    """

    def _pdf(self, x: npt.ArrayLike, sigma: npt.ArrayLike) -> npt.NDArray[np.float64]:
        x = np.asarray(x)
        sigma = np.asarray(sigma)

        return_value: npt.NDArray[np.float64] = (
            np.sqrt(2 / np.pi) / sigma * np.exp(-(x**2) / (2 * sigma**2))
        )
        return return_value


class PairwiseDistance2dIdenticalSigmaZeroMu(stats.rv_continuous):
    """
    A random variable describing the distribution
    of pair distances for two normal distributed point clouds in 2D.

    The continuous distribution class inherits from scipy.stats.rv_continuous
    a set of methods and is defined by overriding the _pdf method.

    For theoretical background see [1]_.

    Parameters
    ----------
    x : npt.ArrayLike
        distance
    sigma : npt.ArrayLike
        Standard deviation for point clouds

    References
    ----------
    .. [1] L. Stirling Churchman, Henrik Flyvbjerg, James A. Spudich,
       A Non-Gaussian Distribution Quantifies Distances Measured with
       Fluorescence Localization Techniques.
       Biophysical Journal 90 (2), 2006, 668-671,
       doi.org/10.1529/biophysj.105.065599.
    """

    def _pdf(self, x: npt.ArrayLike, sigma: npt.ArrayLike) -> npt.NDArray[np.float64]:
        x = np.asarray(x)
        sigma = np.asarray(sigma)

        return_value: npt.NDArray[np.float64] = (
            x / (sigma**2) * np.exp(-(x**2) / (2 * sigma**2))
        )
        return return_value


class PairwiseDistance3dIdenticalSigmaZeroMu(stats.rv_continuous):
    """
    A random variable describing the distribution
    of pair distances for two normal distributed point clouds in 3D.

    The continuous distribution class inherits from scipy.stats.rv_continuous
    a set of methods and is defined by overriding the _pdf method.

    For theoretical background see [1]_.

    Parameters
    ----------
    x : npt.ArrayLike
        distance
    sigma : npt.ArrayLike
        Standard deviation for point clouds

    References
    ----------
    .. [1] L. Stirling Churchman, Henrik Flyvbjerg, James A. Spudich,
       A Non-Gaussian Distribution Quantifies Distances Measured with
       Fluorescence Localization Techniques.
       Biophysical Journal 90 (2), 2006, 668-671,
       doi.org/10.1529/biophysj.105.065599.
    """

    def _pdf(self, x: npt.ArrayLike, sigma: npt.ArrayLike) -> npt.NDArray[np.float64]:
        x = np.asarray(x)
        sigma = np.asarray(sigma)

        return_value: npt.NDArray[np.float64] = (
            np.sqrt(2 / np.pi)
            * x
            / sigma
            * np.exp(-(x**2) / (2 * sigma**2))
            * x
            / (sigma**2)
        )
        return return_value


class _DistributionFits:
    """
    Handle for distribution fits.

    This class is typically instantiated by LocalizationPrecision methods.
    It holds the statistical parameters derived by fitting the result
    distributions using MLE (scipy.stats).

    Parameters
    ----------
    analyis_class : LocalizationPrecision
        The analysis class with result data to fit.

    Attributes
    ----------
    analysis_class : LocalizationPrecision
        The analysis class with result data to fit.
    pairwise_distribution : Pairwise_distance_distribution_2d
        Continuous distribution function used to fit Position_distances
    parameters : list[str]
        Distribution parameters.

    Note
    ----
    Attributes for fit parameter are generated dynamically, named as
    loc_property + distribution parameters and listed in parameters.
    """

    def __init__(self, analysis_class: LocalizationPrecision) -> None:
        self.analysis_class: LocalizationPrecision = analysis_class
        self.distribution: stats.rv_continuous | None = None
        self._dist_parameters: list[str] | None = None
        self.parameters: list[str] = []

        # continuous distributions
        if self.analysis_class.results is None:
            self.pairwise_distribution = None
        else:
            delta_columns = [
                c for c in self.analysis_class.results.columns if "position_delta" in c
            ]
            if len(delta_columns) == 1:
                self.pairwise_distribution = PairwiseDistance1dIdenticalSigmaZeroMu(
                    name="pairwise", a=0.0
                )
            elif len(delta_columns) == 2:
                self.pairwise_distribution = PairwiseDistance2dIdenticalSigmaZeroMu(
                    name="pairwise", a=0.0
                )
            elif len(delta_columns) == 3:
                self.pairwise_distribution = PairwiseDistance3dIdenticalSigmaZeroMu(
                    name="pairwise", a=0.0
                )
            # a is the lower bound of the support of the distribution
            # self.pairwise_distribution = PairwiseDistance2dIdenticalSigma(name='pairwise') also works but is very slow.

    def fit(self, loc_property: str = "position_distance", **kwargs: Any) -> None:
        """
        Fit distributions of results using a MLE fit (scipy.stats) and provide
        fit results.

        Parameters
        ----------
        loc_property
            The property for which to fit an appropriate distribution
        kwargs
            Other parameters passed to the `distribution.fit()` method.
        """
        if self.analysis_class.results is None:
            raise ValueError("Compute results before fitting.")

        # prepare parameters
        if "position_delta_" in loc_property:
            self.distribution = stats.norm
            self._dist_parameters = [
                (loc_property + "_" + param) for param in ["loc", "scale"]
            ]
        elif loc_property == "position_distance":
            self.distribution = self.pairwise_distribution
            self._dist_parameters = [
                (loc_property + "_" + param) for param in ["sigma", "loc", "scale"]
            ]
        else:
            raise TypeError("Unknown localization property.")

        for param in self._dist_parameters:
            if param not in self.parameters:
                self.parameters.append(param)

        # MLE fit of distribution on data
        if "position_delta_" in loc_property:
            fit_results = self.distribution.fit(
                self.analysis_class.results[loc_property].values, **kwargs
            )
        elif loc_property == "position_distance":
            fit_results = self.distribution.fit(
                self.analysis_class.results[loc_property].values,
                **dict(dict(floc=0, fscale=1), **kwargs),
            )
        else:
            raise TypeError("Unknown localization property.")

        for parameter, result in zip(self._dist_parameters, fit_results):
            setattr(self, parameter, result)

    def plot(
        self,
        ax: mpl.axes.Axes | None = None,
        loc_property: str = "position_distance",
        **kwargs: Any,
    ) -> mpl.axes.Axes:
        """
        Provide plot as :class:`matplotlib.axes.Axes` object showing the
        probability distribution functions of fitted results.

        Parameters
        ----------
        ax
            The axes on which to show the image.
        loc_property
            The property for which to plot the distribution fit.
        kwargs
            Other parameters passed to :func:`matplotlib.pyplot.plot`.

        Returns
        -------
        matplotlib.axes.Axes
            Axes object with the plot.
        """
        if ax is None:
            ax = plt.gca()

        if self.distribution is None:
            return ax

        # plot fit curve
        if "position_delta_" in loc_property:
            _loc = getattr(self, loc_property + "_loc")
            _scale = getattr(self, loc_property + "_scale")

            x_values = np.linspace(
                stats.norm.ppf(0.01, loc=_loc, scale=_scale),
                stats.norm.ppf(0.99, loc=_loc, scale=_scale),
                100,
            )
            ax.plot(
                x_values,
                stats.norm.pdf(x_values, loc=_loc, scale=_scale),
                "r-",
                lw=3,
                alpha=0.6,
                label="fitted pdf",
                **kwargs,
            )

        elif loc_property == "position_distance":
            if isinstance(
                self.pairwise_distribution, PairwiseDistance2dIdenticalSigma
            ):  # pragma: no cover
                _sigma = self.position_distance_sigma  # type: ignore
                _mu = self.position_distance_mu  # type: ignore
                x_values = np.linspace(
                    self.pairwise_distribution.ppf(0.01, mu=_mu, sigma=_sigma),
                    self.pairwise_distribution.ppf(0.99, mu=_mu, sigma=_sigma),
                    100,
                )
                ax.plot(
                    x_values,
                    self.pairwise_distribution.pdf(x_values, mu=_mu, sigma=_sigma),
                    "r-",
                    lw=3,
                    alpha=0.6,
                    label="fitted pdf",
                    **kwargs,
                )
            elif isinstance(
                self.pairwise_distribution,
                (
                    PairwiseDistance1dIdenticalSigmaZeroMu,
                    PairwiseDistance2dIdenticalSigmaZeroMu,
                    PairwiseDistance3dIdenticalSigmaZeroMu,
                ),
            ):
                _sigma = self.position_distance_sigma  # type: ignore
                x_values = np.linspace(
                    self.pairwise_distribution.ppf(0.01, sigma=_sigma),
                    self.pairwise_distribution.ppf(0.99, sigma=_sigma),
                    100,
                )
                ax.plot(
                    x_values,
                    self.pairwise_distribution.pdf(x_values, sigma=_sigma),
                    "r-",
                    lw=3,
                    alpha=0.6,
                    label="fitted pdf",
                    **kwargs,
                )
            else:
                raise NotImplementedError(
                    "pairwise_distribution function has not been implemented for plotting "
                    "position distances."
                )

        return ax

    def parameter_dict(self) -> dict[str, float]:
        """Dictionary of fitted parameters."""
        return {k: self.__dict__[k] for k in self.parameters}



"""

Provide colormaps for visualization.

This module provides convenience functions for using colormap definitions
from various visualization packages.

Default colormaps in locan are set through the
:py:data:`locan.configuration.COLORMAP_DEFAULTS` mapping.

Named colormaps are registered through the
:py:data:`locan.visualization.colormap.colormap_registry` mapping.

In locan :py:class:`locan.Colormap` serves as adapter class to provide an
interface for various visualization functions.
Instances of :py:class:`locan.Colormap` can be requested through the
:py:func:`locan.visualization.colormap.get_colormap` function
and contain references to matplotlib and napari colormap instances.

Examples
--------
>>> colormap = locan.get_colormap("viridis")
>>> assert isinstance(colormap.matplotlib, mcolors.Colormap)
>>> colormap.name
viridis

>>> colormap = locan.Colormap.from_matplotlib(colormap="viridis")
>>> assert isinstance(colormap.matplotlib, mcolors.Colormap)
>>> colormap.name
viridis

Variables
----------

.. autosummary::
   :toctree: ./

   colormap_registry

"""

from __future__ import annotations

import logging
import sys
from collections.abc import Mapping
from enum import Enum
from typing import TYPE_CHECKING, Any, TypeVar, Union

if sys.version_info >= (3, 10):
    from typing import TypeAlias
else:
    from typing_extensions import TypeAlias

import matplotlib.pyplot as plt
import numpy as np
from matplotlib import colors as mcolors

from locan.configuration import COLORMAP_DEFAULTS
from locan.dependencies import HAS_DEPENDENCY, needs_package
from locan.visualize.colormap_definitions import (
    cet_coolwarm_colors,
    cet_fire_colors,
    cet_glasbey_dark_colors,
    cet_gray_colors,
)

if HAS_DEPENDENCY["colorcet"]:
    import colorcet

if HAS_DEPENDENCY["napari"]:
    import napari
    import vispy

if TYPE_CHECKING:
    import numpy.typing as npt

logger = logging.getLogger(__name__)

__all__: list[str] = [
    "Colormaps",
    "Colormap",
    "colormap_registry",
    "get_colormap",
]

MatplotlibColormap: TypeAlias = mcolors.Colormap
if HAS_DEPENDENCY["napari"]:
    NapariColormap: TypeAlias = napari.utils.Colormap
    VispyColormap: TypeAlias = vispy.color.Colormap
T_Colormap = TypeVar("T_Colormap", bound="Colormap")

Colormaps = Enum("Colormaps", COLORMAP_DEFAULTS, module=__name__)  # type: ignore[misc]
Colormaps.__doc__ = """
Preferred colormap types to be used for visualization.

Note
----
This enum is automatically generated from COLORMAP_DEFAULTS and should
not be modified.
"""


class Colormap:
    """
    Container class for colormaps.

    A locan Colormap can be instantiated from other colormaps and serves as
    adapter class.
    """

    def __init__(
        self, colormap: MatplotlibColormap | NapariColormap | VispyColormap
    ) -> None:
        self._name: str | None = None
        self._matplotlib: mcolors.Colormap | None = None
        self._napari: napari.utils.Colormap | None = None

        if isinstance(colormap, mcolors.Colormap):
            self._matplotlib = colormap
        elif isinstance(colormap, (napari.utils.Colormap, vispy.color.Colormap)):
            self._napari = colormap
        else:
            raise TypeError(
                "The colormap type is not supported. "
                "Try to use appropriate class methods for construction."
            )

    @property
    def name(self) -> str:
        if self._name is None:
            if self._matplotlib:
                self._name = self._matplotlib.name
            elif self._napari:
                self._name = self._napari.name
            else:
                raise ValueError("No colormap available")
        return self._name

    @property
    def matplotlib(self) -> mcolors.Colormap:
        if self._matplotlib is None:
            if self._napari is not None:
                name = self.napari.name
                colors = self.napari.colors
                controls = self.napari.controls
                self._matplotlib = mcolors.LinearSegmentedColormap.from_list(
                    name, list(zip(controls, colors))
                )
            else:
                raise ValueError("No colormap available")
        return self._matplotlib

    @property
    @needs_package("napari")
    def napari(self) -> napari.utils.Colormap:
        if self._napari is None:
            if self._matplotlib is not None:
                try:
                    self._napari = (
                        napari.utils.colormaps.colormap_utils.vispy_or_mpl_colormap(
                            name=self.matplotlib.name
                        )
                    )
                except KeyError:
                    colors = self(np.linspace(0, 1, 256))
                    colormap_dict = {
                        "colors": colors,
                        "name": self.matplotlib.name,
                        "interpolation": "linear",
                    }
                    self._napari = napari.utils.colormaps.Colormap(**colormap_dict)
            else:
                raise ValueError("No colormap available")
        return self._napari

    @classmethod
    def from_registry(cls: type[T_Colormap], colormap: str) -> Colormap:
        if colormap in colormap_registry:
            return colormap_registry[colormap]
        else:
            raise LookupError("The colormap is not in registry.")

    @classmethod
    def from_matplotlib(
        cls: type[T_Colormap], colormap: str | mcolors.Colormap
    ) -> T_Colormap:
        _matplotlib = plt.get_cmap(name=colormap)
        return cls(colormap=_matplotlib)

    @classmethod
    @needs_package("napari")
    def from_napari(
        cls: type[T_Colormap], colormap: str | dict[str, Any] | NapariColormap
    ) -> T_Colormap:
        if isinstance(colormap, str):
            _napari = napari.utils.colormaps.ensure_colormap(colormap=colormap)
        elif isinstance(colormap, napari.utils.Colormap):
            _napari = colormap
        elif isinstance(colormap, dict):
            _napari = napari.utils.Colormap(**colormap)
        else:
            raise TypeError("Cannot create napari.utils.Colormap from colormap input.")
        return cls(colormap=_napari)

    @classmethod
    @needs_package("colorcet")
    def from_colorcet(cls: type[T_Colormap], colormap: str) -> T_Colormap:
        if isinstance(colormap, str):
            if colormap.startswith("cet_"):
                _matplotlib = mcolors.Colormap(colormap)
            else:
                _matplotlib = colorcet.cm[colormap]
        else:
            raise TypeError("Cannot create Colormap from colormap input.")
        return cls(colormap=_matplotlib)

    def __call__(
        self, *args: Any, **kwargs: Any
    ) -> float | tuple[float, float, float, float] | npt.NDArray[np.float64]:
        if HAS_DEPENDENCY["matplotlib"] and self._matplotlib is not None:
            return_value = self._matplotlib(*args, **kwargs)
        elif HAS_DEPENDENCY["napari"] and self._napari is not None:
            return_value = self._napari.map(*args, **kwargs)
        else:
            raise NotImplementedError("There is no backend for colormaps available.")
        return return_value  # type: ignore


# set up a colormap registry

_colormap_cet_fire = mcolors.LinearSegmentedColormap.from_list(
    name="cet_fire", colors=cet_fire_colors
)

_colormap_cet_gray = mcolors.LinearSegmentedColormap.from_list(
    name="cet_gray", colors=cet_gray_colors
)

_colormap_cet_coolwarm = mcolors.LinearSegmentedColormap.from_list(
    name="cet_coolwarm", colors=cet_coolwarm_colors
)

_colormap_cet_glasbey_dark = mcolors.LinearSegmentedColormap.from_list(
    name="cet_glasbey_dark", colors=cet_glasbey_dark_colors
)

_colormap_registry_from_definitions: dict[str, Colormap] = {
    "cet_fire": Colormap(colormap=_colormap_cet_fire),
    "cet_fire_r": Colormap(colormap=_colormap_cet_fire.reversed()),
    "cet_gray": Colormap(colormap=_colormap_cet_gray),
    "cet_gray_r": Colormap(colormap=_colormap_cet_gray.reversed()),
    "cet_coolwarm": Colormap(colormap=_colormap_cet_coolwarm),
    "cet_glasbey_dark": Colormap(colormap=_colormap_cet_glasbey_dark),
}

_colormap_registry_matplotlib: dict[str, Colormap] = {
    "viridis": Colormap.from_matplotlib("viridis"),
    "viridis_r": Colormap.from_matplotlib("viridis_r"),
    "gray": Colormap.from_matplotlib("gray"),
    "gray_r": Colormap.from_matplotlib("gray_r"),
    "turbo": Colormap.from_matplotlib("turbo"),
    "coolwarm": Colormap.from_matplotlib("coolwarm"),
    "tab20": Colormap.from_matplotlib("tab20"),
}

#: A mapping of names onto Colormap instances.
colormap_registry: Mapping[str, Colormap] = (
    _colormap_registry_from_definitions | _colormap_registry_matplotlib
)


ColormapType = Union[
    str,
    Colormaps,
    Colormap,
    mcolors.Colormap,
    "napari.utils.Colormap",
]


def get_colormap(colormap: ColormapType) -> Colormap:
    """
    Get a locan.Colormap instance from colormap searching string identifier through
    colormap_registry, matplotlib colormaps, napari_colormaps.

    Parameters
    ----------
    colormap
        Colormap request

    Returns
    -------
    Colormap
    """
    if isinstance(colormap, Colormap):
        return colormap
    elif isinstance(colormap, mcolors.Colormap):
        return Colormap.from_matplotlib(colormap=colormap)
    elif isinstance(colormap, Colormaps):
        return Colormap.from_registry(colormap.value)
    elif HAS_DEPENDENCY["napari"] and isinstance(colormap, napari.utils.Colormap):
        return Colormap.from_napari(colormap=colormap)
    elif isinstance(colormap, str):
        try:
            return Colormap.from_registry(colormap)
        except LookupError:
            pass
        try:
            return Colormap.from_matplotlib(colormap=colormap)
        except ValueError:
            pass
        if HAS_DEPENDENCY["napari"]:
            try:
                return Colormap.from_napari(colormap=colormap)
            except KeyError:
                pass
        raise TypeError(
            f"The colormap {colormap} is not available in either colormap_registry, "
            f"matplotlib or napari."
        )
    else:
        raise TypeError("No such colormap available.")



from locan import dependency_info, show_versions, system_info


def test__get_sys_info():
    sys_info = system_info()
    # print(sys_info)
    assert len(sys_info) == 12


def test___get_dependency_info():
    deps_info = dependency_info(other_dependencies=["locan"])
    # print(deps_info)
    assert "boost-histogram" in deps_info
    assert "locan" in deps_info


def test_show_versions_(capfd):
    show_versions(other_dependencies=["locan"])
    captured = capfd.readouterr()
    for header in [
        "\nLocan:",
        "\nPython:",
        "\nSystem:",
        "\nPython dependencies:",
        "locan",
        "node:",
        "executable:",
    ]:
        assert header in captured.out

    show_versions(verbose=False, extra_dependencies=False)
    captured = capfd.readouterr()
    for header in ["\nLocan:", "\nPython:", "\nSystem:", "\nPython dependencies:"]:
        assert header in captured.out
    for header in ["node:", "executable:"]:
        assert header not in captured.out



import os
import warnings

import locan
from locan.__main__ import main


def test_version():
    try:
        assert isinstance(locan._version.version, str)
        assert isinstance(locan._version.version_tuple, tuple)
    except AttributeError:
        warnings.warn(
            "AttributeError was raised; "
            "most likely because locan._version is not available.",
            stacklevel=2,
        )
    assert locan.__version__
    intersection = set(dir(locan)).difference(set(locan.__all__))

    # print(locan.__version__)
    # print(len(dir(locan)))  # currently: 211
    # print(dir(locan))
    # print(len(locan.__all__))
    # print(locan.__all__)
    # print(intersection)

    assert not intersection


def test_entrypoint(capfd):
    main([])
    captured = capfd.readouterr()
    assert captured.out[:47] == "This is the command line entry point for locan."


def test_entrypoint_from_sys(capfd):
    exit_status = os.system("locan")
    captured = capfd.readouterr()
    assert captured.out[:47] == "This is the command line entry point for locan."
    assert exit_status == 0



#!/usr/bin/env python

"""
Run project test suite with pytest.


To run the script::

    test

Try for instance::

    test

See Also
--------
locan.tests.test
"""
from __future__ import annotations

import argparse

from locan.tests import test as sc_test


def _add_arguments(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "pytest_args",
        nargs="?",
        type=str,
        default=None,
        help="string with pytest options.",
    )


def main(args: list[str] | None = None) -> None:
    parser = argparse.ArgumentParser(description="Run test suite.")
    _add_arguments(parser)
    returned_args = parser.parse_args(args)
    sc_test(args=returned_args.pytest_args)


if __name__ == "__main__":
    main()
